# 2021 Leo Kaindl

import json
import logging
import random
import re
import string
import threading
import webbrowser
from time import sleep

import gi
import pandas as pd
import requests
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from ab12phylo import repo
from ab12phylo_cmd.filter import new_id
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
PAGE = 1
ERRORS = ['groups?', 'no match', 'use groups!']


class rgx_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        iface.single_rt.connect('toggled', self._switch_single_or_triple_regex, )
        iface.triple_rt.connect('toggled', self._switch_single_or_triple_regex)
        iface.triple_rt.join_group(iface.single_rt)
        iface.single_rt.set_active(True)

        iface.regex_apply.connect('clicked', self.parse_all, True)
        iface.wp_apply.connect('clicked', self.parse_single_group, iface.wp_rx, 2)
        iface.single_rx.connect('activate', self.parse_all, True)

        for thing, col in zip(['well', 'gene', 'plate', 'wp'], [2, 4, 3, 2]):
            entry = iface.__getattribute__('%s_rx' % thing)
            entry.connect('activate', self.parse_single_group, entry, col)
            entry.connect('focus_out_event', self.parse_single_group, entry, col)

        iface.view_trace_regex.set_model(data.trace_store)
        iface.view_csv_regex.set_model(data.plate_store)
        iface.reverse_rx_chk.set_active(data.search_rev)
        iface.reverse_rx_chk.connect('toggled', self._switch_search_reverse_reads)
        iface.reverse_rx.connect('activate', self.parse_single_group, iface.reverse_rx, 6)
        iface.regex_try_online.connect('clicked', self.try_online)

        # allow deletion
        for widget, tv in zip([iface.remove_path_regex, iface.remove_csv_regex],
                              [iface.view_trace_regex, iface.view_csv_regex]):
            sel = tv.get_selection()
            sel.set_mode(Gtk.SelectionMode.MULTIPLE)
            widget.connect('clicked', self.delete_files_from_input_selection, PAGE, sel)
            tv.connect('key-press-event', self.tv_keypress, PAGE, sel)

        self.reset_columns()

    def reset_columns(self, do_parse=False):
        data = self.data
        iface = self.iface
        self.check_for_plates()

        # remove old columns:
        for tv in [iface.view_trace_regex, iface.view_csv_regex]:
            [tv.remove_column(col) for col in tv.get_columns()]

        # columns depend on if we are reading plates
        if iface.plates:
            for title, j in zip(['well', 'plate', 'gene', 'file'], [2, 3, 4, 1]):
                crt = Gtk.CellRendererText(editable=True)
                crt.connect('edited', self.cell_edit, iface.view_trace_regex, j)
                col = Gtk.TreeViewColumn(title=title, cell_renderer=crt, text=j, foreground=7)
                col.set_sort_column_id(j)
                iface.view_trace_regex.append_column(col)
            # wellsplates:
            for title, j in zip(['plate ID', 'file'], [2, 1]):
                crm = Gtk.CellRendererText(editable=True)
                crm.connect('edited', self.cell_edit, iface.view_csv_regex, j)
                col = Gtk.TreeViewColumn(title=title, cell_renderer=crm, text=j, foreground=3)
                col.set_sort_column_id(j)
                iface.view_csv_regex.append_column(col)
        else:
            for title, j in zip(['sample', 'gene', 'file'], [2, 4, 1]):
                crt = Gtk.CellRendererText(editable=True)
                crt.connect('edited', self.cell_edit, iface.view_trace_regex, j)
                col = Gtk.TreeViewColumn(title=title, cell_renderer=crt, text=j, foreground=7)
                col.set_sort_column_id(j)
                iface.view_trace_regex.append_column(col)
            # set regex for plate_id column as already fired
            data.rx_fired[-2] = True

        self._switch_search_reverse_reads(self.iface.reverse_rx_chk)
        data.rx_fired[6] = True

        if do_parse:  # fire the initial parse
            self.parse_all(None, force=True)
            self.refresh(page=1)

    def parse_single_group(self, widget, entry, col, fourth=None, refresh_after=True):
        if fourth:  # caught a focus_out_event, therefore shifted arguments
            entry, col = col, fourth
        data = self.data
        iface = self.iface
        LOG.debug('single')
        try:
            regex = re.compile(entry.get_text())
        except re.error:
            self.show_notification('RegEx could not be compiled.')
            return
        errors = self.get_errors(PAGE)
        changed = self.get_changed(PAGE)

        # check if traces or plates
        if widget in [iface.wp_rx, iface.wp_apply]:
            model = data.plate_store
            traces = False
            data.rx_fired[col + data.trace_store.get_n_columns()] = True
        else:
            model = data.trace_store
            traces = True
            data.rx_fired[col] = True

        # parse each filename
        for i, row in enumerate(model):
            # skip references
            if traces and row[5]:
                errors = errors or row[-1] == iface.AQUA
                continue

            file = row[1]
            if entry is not iface.reverse_rx:
                try:
                    m = regex.search(file).groups()[0]
                    if not changed:
                        changed = m != row[col]
                    model[i][col] = m
                    model[i][-1] = None  # iface.FG
                    continue
                except ValueError as ve:
                    # maybe wrong number of groups
                    model[i][col] = ERRORS[0]
                except AttributeError as ae:
                    # no match
                    model[i][col] = ERRORS[1]
                except IndexError as ie:
                    # no groups used
                    model[i][col] = ERRORS[2]
                model[i][-1] = iface.RED
                errors, changed = True, True
            else:
                # reverse reads, save boolean values
                try:
                    m = regex.search(file).groups()[0]
                    # reverse read
                    if not changed:
                        changed = row[col] is not True
                    model[i][col] = True
                    if model[i][-1] != iface.RED:
                        model[i][-1] = None  # iface.FG
                except AttributeError as ae:
                    # forward read
                    if not changed:
                        changed = row[col] is not False
                    model[i][col] = False
                    if model[i][-1] != iface.RED:
                        model[i][-1] = None  # iface.FG
                except (IndexError, ValueError):
                    errors, changed = True, True
                    model[i][-1] = iface.RED
                except Exception as ex:
                    raise RuntimeWarning from ex
        self.set_changed(PAGE, changed)
        self.set_errors(PAGE, errors)
        if errors and sum(data.rx_fired) >= 5 and refresh_after:
            self.refresh()

    def parse_all(self, wi, force=False):
        data = self.data
        iface = self.iface
        if not force:  # it's easier to pass an additional param than blocking handlers
            return

        if iface.single_rt.get_active():
            LOG.debug('triple')
            try:
                regex = re.compile(iface.single_rx.get_text())
            except re.error:
                self.show_notification('RegEx could not be compiled')
                return
            errors = self.get_errors(PAGE)
            changed = self.get_changed(PAGE)

            for idx, row in enumerate(data.trace_store):
                # skip references
                if row[5]:  # is reference
                    errors = errors or row[-1] == iface.AQUA
                    continue
                file = row[1]
                try:
                    m = regex.search(file)
                    plate, gene, well = m.groups() if iface.plates else (None, m.groups())
                    if not changed:
                        changed = not bool(row[2:5] == [well, plate, gene])
                    data.trace_store[idx] = row[:2] + [well, plate, gene] + row[5:]
                    data.trace_store[idx][-1] = None  # iface.FG
                    continue
                except ValueError as ve:
                    data.trace_store[idx][2:5] = [ERRORS[0], '', '']
                except AttributeError as ae:
                    data.trace_store[idx][2:5] = [ERRORS[1], '', '']
                except IndexError as ie:
                    data.trace_store[idx][2:5] = [ERRORS[2], '', '']
                data.trace_store[idx][7] = iface.RED
                errors, changed = True, True
            self.set_changed(PAGE, changed)
            self.set_errors(PAGE, errors)
            data.rx_fired[2:5] = [True] * 3
        else:
            self.parse_single_group(None, iface.well_rx, 2, False)
            self.parse_single_group(None, iface.gene_rx, 4, False)
        if iface.plates:
            self.parse_single_group(iface.wp_apply, iface.wp_rx, 2, False)
        if iface.reverse_rx_chk.get_active():
            self.parse_single_group(None, iface.reverse_rx, 6, False)
        if self.get_errors(PAGE) and sum(data.rx_fired) >= 5:
            self.refresh()

    def cell_edit(self, cell, path, new_text, tv, col):
        self.save_row_edits(cell, path, new_text, tv, col)
        self.set_changed(PAGE)
        # set row to error-free. also for references
        if tv == self.iface.view_trace_regex:
            mo = tv.get_model()
            mo[path][-1] = self.iface.BLUE if mo[path][5] else None  # iface.FG
        self.refresh()

    def _switch_single_or_triple_regex(self, widget):
        # (In)activates the Entry fields and causes a parse event
        data = self.data
        iface = self.iface
        if widget.get_active():
            self.set_errors(PAGE, False)  # important, do not pre-suppose it will fail
            three = ['plate_rx', 'gene_rx', 'well_rx', 'plate_regex_label', 'gene_regex_label', 'well_regex_label']
            if widget == iface.single_rt:
                iface.single_rx.set_sensitive(True)
                [iface.__getattribute__(widget).set_sensitive(False) for widget in three]
            else:
                iface.single_rx.set_sensitive(False)
                [iface.__getattribute__(widget).set_sensitive(True) for widget in three]
            # adjust the sensitivity of the plate regex line
            if not iface.plates:
                iface.plate_rx.set_sensitive(False)
                iface.plate_regex_label.set_sensitive(False)
            LOG.debug('toggled radiobutton %s' % widget.get_name())
            # parse again
            self.parse_all(widget)

    def _switch_search_reverse_reads(self, widget):
        data = self.data
        iface = self.iface
        n_cols = iface.view_trace_regex.get_n_columns()
        if widget.get_active():
            # save in data
            data.search_rev = True
            # enable entry field
            iface.reverse_rx.set_sensitive(True)
            # create new column
            crt = Gtk.CellRendererToggle(radio=False)
            crt.connect('toggled', self._manual_mark_as_reversed)
            col = Gtk.TreeViewColumn(title='rev', cell_renderer=crt, active=6)
            col.set_sort_column_id(6)
            iface.view_trace_regex.insert_column(col, n_cols - 1)
            # set regex as not yet fired
            data.rx_fired[6] = False
            # cause parsing
            self.parse_single_group(None, iface.reverse_rx, 6)
        else:
            # save in data
            data.search_rev = False
            iface.reverse_rx.set_sensitive(False)
            for col in iface.view_trace_regex.get_columns():
                if col.get_title() == 'rev':
                    iface.view_trace_regex.remove_column(col)
            # set regex as already fired
            data.rx_fired[6] = True

    def _manual_mark_as_reversed(self, widget, path):
        """Mark a record as reverse or not"""
        data = self.data
        data.trace_store.set(data.trace_store.get_iter(path), [6], [not data.trace_store[path][6]])
        self.set_changed(PAGE)
        self.refresh()

    def check_for_plates(self):
        iface = self.iface
        plates_now = len(self.data.plate_store) > 0
        if 'plates' not in iface or plates_now != iface.plates:
            iface.plates = plates_now

            # close and inactivate the expander
            iface.wp_expander.set_expanded(iface.plates)
            iface.wp_expander.set_sensitive(iface.plates)
            self.reset_columns()

        if not iface.plates:
            # inactivate the line for the matching single regex
            if not iface.triple_rt.get_active():
                [iface.__getattribute__(name).set_sensitive(False) for name in
                 ['plate_rx', 'plate_regex_label']]

    def refresh(self, **kwargs):
        data = self.data
        iface = self.iface
        self.check_for_plates()
        # make all columns sortable
        for tv in [iface.view_trace_regex, iface.view_csv_regex]:
            tv.columns_autosize()

        # try matching reference files to genes
        data.genes = sorted({gene for gene, is_ref, color in data.trace_store.get_column((4, 5, 7))
                             if not is_ref and color is not iface.RED and gene != ''})
        if not data.genes or data.genes == ['']:
            iface.gene_list.set_label('no genes')
            iface.gene_list.set_visible(False)
            return
        tx = ', '.join(data.genes)
        LOG.debug(f'genes: {tx}')
        iface.gene_list.set_label(tx)
        iface.gene_list.set_visible(True)
        single_gene = data.genes[0] if len(data.genes) == 1 else False

        for i, row in enumerate(data.trace_store):
            # iterate over reference sequences and mark them blue if they are ok
            if row[5] and row[-1] == iface.AQUA:
                if single_gene:
                    data.trace_store[i][4] = single_gene
                    data.trace_store[i][7] = iface.BLUE
                else:
                    file_name = row[1].upper()
                    for gene in data.genes:
                        if gene.upper() in file_name:
                            data.trace_store[i][4] = gene
                            data.trace_store[i][7] = iface.BLUE
                            break

        # check the dataset for red lines
        trace_colors = data.trace_store.get_column(-1)
        if iface.RED not in trace_colors and iface.AQUA not in trace_colors \
                and iface.RED not in data.plate_store.get_column(-1):
            self.set_errors(PAGE, False)
            data.rx_fired[2:5] = [True] * 3
            data.rx_fired[6] = True
        else:
            assert self.get_errors(PAGE) or sum(data.rx_fired) < 5

        # check the dataset for versionized entries
        appears = dict()
        for w, p, g in data.trace_store.get_column((2, 3, 4)):
            if w not in appears:
                appears[w] = {p: {g: 1}}
            elif p not in appears[w]:
                appears[w][p] = {g: 1}
            else:
                appears[w][p][g] = appears[w][p].get(g, 0) + 1

        for row in data.trace_store:
            if appears[row[2]][row[3]][row[4]] > 1:
                row[-1] = iface.PURPLE

    def reload_ui_state(self):
        for w_name in ['single_rx', 'well_rx', 'gene_rx', 'plate_rx', 'reverse_rx', 'wp_rx']:
            self.iface.__getattribute__(w_name).set_text(str(self.data.rgx.__getattribute__(w_name)))
        for w_name in ['single_rt', 'triple_rt', 'reverse_rx_chk']:
            self.iface.__getattribute__(w_name).set_active(self.data.rgx.__getattribute__(w_name))

    def try_online(self, *args):
        data = self.data
        iface = self.iface
        LOG.debug('generating online help')
        url = 'https://regex101.com/api/regex/'
        headers = {'content-type': 'application/json'}

        payload = dict(flavor='python', flags='gm', delimiter='"')
        payload['regex'] = iface.single_rx.get_text()
        content = tuple(i.get_text() for i in [iface.wp_rx, iface.single_rx, iface.well_rx,
                                               iface.gene_rx, iface.plate_rx, iface.reverse_rx]) \
                  + ('\n'.join([row[1] for row in data.plate_store]),
                     '\n'.join([row[1] for row in data.trace_store]))

        payload['testString'] = 'wellsplate ID:\n%s\n\nsingle RegEx:\n%s\n\n' \
                                'separate expressions:\n%s\n%s\n%s\n\n' \
                                'reverse read:\n%s\n\n' \
                                'plate filenames:\n%s\n\n' \
                                'trace filenames:\n%s\n' % content

        payload = json.dumps(payload, indent=2)
        try:
            response = requests.request("POST", url, data=payload, headers=headers)
            result = json.loads(response.text)

            webbrowser.open('https://regex101.com/r/' + result['permalinkFragment'])

            # wait for the page to load
            sleep(4)
            # delete the entry
            payload = json.dumps(dict(deleteCode=result['deleteCode']), indent=2)
            response = requests.request("DELETE", url, data=payload, headers=headers)
        except Exception:
            self.show_notification('online help failed')

    def start_read(self, run_after=None):
        """Called from REFRESH button and shared.proceed"""
        data = self.data
        iface = self.iface

        # stop if there are no traces
        if len(data.trace_store) == 0:
            self.show_notification('No sequence data!')
            return
        if run_after is None:
            self.refresh()  # try matching reference files to genes

        # save_ui_state
        for w_name in ['single_rx', 'well_rx', 'gene_rx', 'plate_rx', 'reverse_rx', 'wp_rx']:
            data.rgx.__setattr__(w_name, iface.__getattribute__(w_name).get_text())
        for w_name in ['single_rt', 'triple_rt', 'reverse_rx_chk']:
            data.rgx.__setattr__(w_name, iface.__getattribute__(w_name).get_active())

        iface.thread = threading.Thread(target=self.do_read)
        iface.run_after = run_after
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()
        # return to main loop

    def do_read(self):
        data = self.data
        iface = self.iface
        iface.text = ''
        iface.i = 0
        iface.k = len(data.trace_store) + len(data.plate_store) + 2
        data.csvs.clear()
        data.seqdata.clear()
        data.record_order.clear()
        records, warnings, errors = list(), set(), list()

        # use random but reproducible prefixes for references. different seed from raxml -> less confusion
        random.seed(data.seed + 2)

        # dict of reference source organism -> random key
        lookup = dict()

        # read in wellsplates
        LOG.debug('reading wellsplates')
        iface.text = 'reading plates ...'
        for row in data.plate_store:
            df = pd.read_csv(row[0], header=None, engine='python')
            df.index = list(range(1, df.shape[0] + 1))
            df.columns = list(string.ascii_uppercase[0:df.shape[1]])
            file, box = row[1:3]
            if box in data.csvs:
                errors.append('overwrite wellsplate %s with %s' % (box, file))
            data.csvs[box] = df.to_dict()
            iface.i += 1

        # read in FASTAs containing more than one record first
        fasta_paths = {file_path.split('~~')[0] for file_path
                       in data.trace_store.get_column(0) if '~~' in file_path}
        fasta_records = {file_path: {r.id: r for r in SeqIO.parse(
            file_path, 'fasta')} for file_path in fasta_paths}

        # read in trace files
        LOG.debug('reading traces')
        found_numerical = False
        for row in data.trace_store:
            file_path, file, coords, box, gene, is_ref, is_rev, color = row
            iface.text = 'reading %s' % file

            # read in one file
            if file_path.endswith('.ab1') and '~~' not in file_path:
                try:
                    records = [SeqIO.read(file_path, 'abi')]  # ABI traces also only contain a single record!
                except UnicodeDecodeError:
                    errors.append('ABI trace error %s' % file)
            elif '~~' in file_path:
                file_part, record_name = file_path.split('~~')
                records = [fasta_records[file_part][record_name]]
            else:
                try:
                    records = SeqIO.parse(file_path, 'fasta')
                except UnicodeDecodeError:
                    errors.append('Seq file error %s' % file)

            # any kind of seq can define a new gene
            if gene not in data.seqdata:
                data.seqdata[gene] = dict()
                data.metadata[gene] = dict()
                data.genes = sorted(set(data.genes) | {gene})
            keys = list(data.seqdata[gene].keys()) if gene in data.seqdata else list()

            # iterate over records found in file (usually only one)
            for record in records:
                if not is_ref:  # normal sequence
                    try:
                        # swap out well coordinates for isolate numbers
                        (y, x) = (int(coords[1:]), coords[0])
                        _id = data.csvs[box][x][y]
                    except (KeyError, ValueError):
                        if len(data.csvs) == 0:
                            # record.id = record.name.replace(gene, '')
                            if box in ['', ' ', '-', '_']:
                                _id = coords.upper()
                            else:
                                _id = box + '_' + coords
                        else:
                            _id = record.id
                            errors.append('missing %s on wellsplate %s' % (coords, box))
                    attributes = {'file': file_path, 'box': box, 'is_rev': is_rev}

                    if _id.isdigit():
                        # prepend a T to numerical IDs so toytree won't get confused
                        _id = f'T{_id}'
                        found_numerical = True
                    if _id in keys:
                        # add suffix to duplicate IDs
                        warnings.add('duplicate ID %s' % _id)
                        _id = new_id(_id, keys)

                else:  # reference
                    # parse species and possibly strain
                    strain = re.split(r'[\s_]', record.description.strip().split(',')[0])
                    accession = strain.pop(0)
                    species = strain.pop(0) + ' ' + strain.pop(0)
                    try:
                        ix = strain.index('strain')
                        strain = ' '.join(strain[ix + 1:ix + 3])
                        # cope with longer composite species names
                        if ix > 0:
                            species += ' ' + ' '.join(strain[0:ix])
                    except ValueError:
                        strain = species

                    # catch some illegal weird html characters
                    for char in ['<', '>', '\'', '"', '&']:
                        strain = strain.replace(char, '')

                    # retrieve or generate random key
                    if strain in lookup:
                        _id = lookup[strain]
                    else:
                        # swap original ids for random short ones that no tool takes offense at.
                        _id = 'REF_' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=5))
                        lookup[strain] = _id

                    # save original id+description
                    attributes = {'file': file_path, 'accession': accession, 'is_rev': is_rev,
                                  'reference_species': species if strain == species
                                  else species + ' strain ' + strain}

                if is_rev:  # strange if it's a reference, but possible anyway
                    record = record.reverse_complement(record.id)
                record = SeqRecord(record.seq, _id, description='', name='',
                                   letter_annotations=getattr(record, 'letter_annotations'))
                # save SeqRecord
                data.seqdata[gene][_id] = record
                data.metadata[gene][_id] = attributes
                # save the order of all records
                data.record_order.append([_id, gene])

            iface.i += 1

        if found_numerical:
            warnings.add('Found records with purely numerical IDs; prepended a T to those.')
        iface.text = 'search missing samples'
        LOG.debug(iface.text)
        data.gene_ids = {g: set(gd.keys()) for g, gd in data.seqdata.items() if g in data.genes}
        all_ids = set.union(*data.gene_ids.values())
        missing = {g: ', '.join(sorted(all_ids - ids))
                   for g, ids in data.gene_ids.items() if all_ids - ids}
        if missing:
            with open(self.wd / repo.PATHS.missing, 'w') as fh:
                fh.write('gene\tmissing samples\n')
                for k, v in missing.items():
                    fh.write('%s\t%s\n' % (k, v))
                    warnings.add('%s missing %s' % (k, v))

        iface.i += 1
        iface.text = 'writing metadata'
        LOG.debug(iface.text)
        self.write_metadata()

        iface.text = 'idle'
        iface.frac = 1
        LOG.debug('reading done')
        GObject.idle_add(self.stop_read, errors, warnings)
        return

    def stop_read(self, errors, warnings):
        """
        Join a thread, then call one or more functions *afterwards*.
        Usually with the same args: gui. These params are stored in iface.
        :param errors:
        :param warnings:
        :return:
        """
        self.iface.thread.join()
        LOG.info('rgx thread idle')
        if errors or warnings:
            self.show_notification('File troubles', ['error: %s' % f for f in errors] +
                                   ['warning: %s' % f for f in warnings], secs=0)
        self.set_changed(PAGE, False)

        # *Now* go back to where you came from
        if self.iface.run_after:
            [do_func() for do_func in self.iface.run_after]
        return



# 2021 Leo Kaindl
import argparse
import logging
from argparse import Namespace

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

LOG = logging.getLogger(__name__)


class project_dataset:
    def __init__(self):
        self.trace_store = picklable_liststore(str,  # path
                                               str,  # filename
                                               str,  # well/id
                                               str,  # plate
                                               str,  # gene
                                               bool,  # reference
                                               bool,  # reversed
                                               str)  # color

        self.plate_store = picklable_liststore(str,  # path
                                               str,  # filename
                                               str,  # plate ID
                                               str)  # errors
        self.rx_fired = [False] * 12
        # rx_fired stores if these columns have been parsed before:
        # [bool] with trace and then plate columns. sum(iface.rx_fired) = 5
        self.genes = list()  # used also before seqdata exists
        self.csvs = dict()
        self.seqdata = dict()
        self.metadata = dict()
        self.seed = 0
        self.record_order = list()
        self.qal_model = picklable_liststore(str,  # id
                                             str,  # gene
                                             int,  # has phreds
                                             bool)  # low quality
        self.search_rev = False
        self.rgx = Namespace()
        self.qal = Namespace(gene_roll='all', accept_rev=True, accept_nophred=True,
                             min_phred=30, trim_out=8, trim_of=10, bad_stretch=5)
        self.msa = Namespace(cmd=dict(), remote_cmd=dict(), last_cmd='')
        self.gbl = Namespace(ignore_ids=set())
        self.ml = Namespace(evo_model='GTR', evo_modify='+G4', cpu_count='1', cpu_use='1',
                            tool='raxml-ng', ml_cmd='command not yet defined')
        self.phy = Namespace(gap_share=.1, unk_share=.1, flip=.7, dist=0, space=1, sel_gene=None,
                             did_BLAST=False, query='', exclude='',
                             rect=True, circ=False, unro=False, tbe=True, fbp=False, supp=True,
                             spec=True, axis=True, align=False, pmsa=False,
                             pdf=False, svg=False, png=True, nwk=True, html=False,
                             h=dict(), qh=dict(), modify=dict())

        self.gbl_model = picklable_liststore(str)  # id
        # set up indicator of changes, tabs are not disabled initially
        self.change_indicator = [False] * 20
        self.errors_indicator = [False] * 20
        self.page = 0
        self.qal_shape = [0, 0]
        self.gbl_shape = [0, 0]  # width-height
        self.msa_shape = [0, 0, 0, 0]  # width-height before and after trimming
        self.msa_lens = list()
        self.msa_hash = ''
        self.gene_ids = dict()  # {gene: {seq_ids}}
        self.gene_for_preview = ''
        self.sp_model = picklable_liststore(str,  # id
                                            str,  # pid
                                            str,  # species
                                            str,  # extra_species
                                            str)  # low quality indicator color
        self.remote_dbs = picklable_liststore(str,  # name
                                              int)  # for synchronization with ComboBox, i think ...
        self.blast_path = None  # for non-$PATH BLAST+ executable
        self.evo_models = picklable_liststore(str, str)  # id, use_file
        [self.evo_models.append([evo, None]) for evo in ['GTR', 'JC', 'SYM', 'K80', 'K81', 'HKY', 'TN93',
                                                         'TPM2', 'TPM3', 'TIM2', 'TIM3', 'TVM', 'from file']]
        self.pop_model = picklable_liststore(str,  # gene
                                             int,  # valid sites
                                             int,  # S
                                             float,  # k
                                             float,  # pi
                                             float,  # theta_w
                                             float,  # D_T
                                             int,  # unique seqs
                                             int,  # gaps
                                             int)  # unknown
        self.tree_anno_model = picklable_liststore(str,  # sample ID
                                                   str,  # species
                                                   str,  # foreground color
                                                   str)  # background color
        self.colors = None

    def new_project(self):
        self.overwrite(project_dataset())

    def overwrite(self, new_dataset):
        for attr in [a for a in dir(self) if not callable(getattr(self, a)) and not a.startswith('__')]:
            old = self.__getattribute__(attr)
            if type(old) == picklable_liststore:
                old.clear()
                try:
                    [old.append(row[:]) for row in new_dataset.__getattribute__(attr)]
                except (AttributeError, ValueError):
                    pass  # try to go without this model
            elif type(old) == dict:
                old.clear()
                try:
                    old.update(new_dataset.__getattribute__(attr))
                except (TypeError, AttributeError):
                    pass  # try to go without dict values
            elif type(old) == argparse.Namespace:
                try:
                    for k, v in new_dataset.__getattribute__(attr).__dict__.items():
                        old.__setattr__(k, v)
                except (AttributeError, TypeError, ValueError):
                    pass  # let's be courageous
            else:
                try:
                    self.__setattr__(attr, new_dataset.__getattribute__(attr))
                except (AttributeError, TypeError) as ex:
                    LOG.error(ex)


class picklable_liststore(Gtk.ListStore):
    """kudos go to samplebias on https://stackoverflow.com/a/5969700"""

    def __init__(self, *args):
        Gtk.ListStore.__init__(self, *args)
        self.col_types = args

    def __reduce__(self):
        rows = [list(row) for row in self]
        return _unpickle_liststore, (self.__class__, self.col_types, rows)

    def get_column(self, col_idx):
        """Extract a column from a Gtk.ListStore, because they are annoying"""
        col = list()
        for row in self:
            col.append(row[col_idx])
        return col


def _unpickle_liststore(cls, col_types, rows):
    inst = cls.__new__(cls)
    inst.__init__(*col_types)
    for row in rows:
        inst.append(row)
    return inst



# 2021 Leo Kaindl

import configparser
import copy
import os
import re
import shutil
import stat
import sys
import tarfile
import zipfile
from pathlib import Path
from subprocess import run, check_output, PIPE

import requests
from Bio.Seq import MutableSeq
from bs4 import BeautifulSoup

regex = re.compile('\\.[\\d]+$')


def trim_ends(seqrecord, min_phred, end_ratio, trim_preview=False):
    """
    Trims a supplied SeqRecord with given minimal quality score and ratio of good bases in end of given length.
    """
    if not seqrecord.letter_annotations:
        raise AttributeError('no quality')
    phreds = seqrecord.letter_annotations['phred_quality']
    if set(phreds) == {0} and min_phred > 0:
        raise AttributeError('no quality')

    # trim left
    start = _ok = 0
    while _ok < end_ratio[0] and start < len(phreds):
        Ns = len([i for i in phreds[start: start + end_ratio[1]] if i < min_phred])
        # assume all good bases are at right edge of counting window
        start += Ns
        # and allow for some leading non-confident bases
        start -= (end_ratio[1] - end_ratio[0])
        _ok = end_ratio[1] - Ns
    start = max(0, start)

    # trim right
    _ok = 0
    end = len(phreds)
    while _ok < end_ratio[0] and end >= 0:
        Ns = len([i for i in phreds[end - end_ratio[1]: end] if i < min_phred])
        end += -Ns + end_ratio[1] - end_ratio[0]
        _ok = end_ratio[1] - Ns

    if start >= end:
        raise ValueError('low quality')

    if trim_preview:
        # make some copies
        qal = copy.copy(phreds)
        rec = copy.copy(seqrecord)
        rec.letter_annotations = copy.copy(rec.letter_annotations)
        rec.letter_annotations['phred_quality'] = qal

        if type(rec.seq) != MutableSeq:
            rec.seq = rec.seq.tomutable()

        if start > 0:
            rec.seq[0:start] = '-' * start
            qal[0:start] = [min_phred + 1] * start

        assert end > 0
        rec.seq[end:] = '-' * (len(rec) - end)
        qal[end:] = [min_phred + 1] * (len(rec) - end)
        return rec
    else:
        return seqrecord[start:end]


def mark_bad_stretches(seqrecord, min_phred, _len):
    """
    Replaces stretches of bases that were sequenced with below-minimal phred quality score in a
    :class:`Bio.SeqRecord` object by a sequence of undetermined (:code:`N`) nucleotides of equal length.
    *Bad stretches* shorter than :code:`_len` are not changed.

    :param seqrecord: The `SeqRecord` object that will be filtered.
    :param min_phred: The minimal score below with replacement with ``N`` might occur
    :param _len: The minimal length of bad bases that will be replaced.
    :return: The input ``SeqRecord``, with bad stretches replaced.
    """

    if not seqrecord.letter_annotations:
        raise AttributeError('no quality')
    phreds = seqrecord.letter_annotations['phred_quality']
    if type(seqrecord.seq) != MutableSeq:
        seqrecord.seq = seqrecord.seq.tomutable()
    start = 0

    while True:
        # look for bad base
        if phreds[start] < min_phred:
            end = start + 1
            while end < len(phreds) and phreds[end] < min_phred:
                # elongate bad stretch
                end += 1
            if end - start >= _len:
                # replace long bad stretch
                seqrecord.seq[start: end] = 'N' * (end - start)
            start += 1
        else:
            # look elsewhere
            start += 1
        if start >= len(seqrecord):
            break

    return seqrecord


def new_id(rid, keys):
    while rid in keys:
        match = regex.search(rid)
        if match:
            rid = rid[:match.start()] + '.' \
                  + str(int(match.group()[1:]) + 1)
        else:
            rid += '.1'
    return rid


def mark_bad_bases(seqrecord, min_phred):
    """
    unused: mark *all* non-confident bases as N
    """
    phreds = seqrecord.letter_annotations['phred_quality']
    seq = str()
    for pos in range(len(phreds)):
        if phreds[pos] < min_phred:
            seq += 'N'
        else:
            seq += seqrecord.seq[pos]
    seqrecord.seq = MutableSeq(seq)


def chmod_x(p):
    """Make a file executable. Accepts either a string or a pathlib.Path"""
    if type(p) == str:
        p = Path(p).resolve()
    try:
        p.chmod(p.stat().st_mode | stat.S_IEXEC)
    except FileNotFoundError:
        pass


def fetch_non_python_tools(suffix, cfg, other_cfg, save_dir, log):
    """
    :param suffix: Indicator it this is ab12phylo -> '' or ab12phylo-cmd -> '-cmd'
    :param cfg: Path to the 'conf.cfg' file
    :param other_cfg: Path to the 'conf.cfg' file for the other AB12PHYLO version
    :param save_dir: Installation target directory. Actually always ./tools, where . is where this file is
    :param log: logging.Logger instance
    :return:
    """
    paths = dict()
    all_tools = ['blastn', 'raxml-ng', 'iqtree2']
    skip = False

    def find(tool):
        try:
            exe = next(save_dir.rglob(f'{tool}{os.getenv("PATHEXT", default="")}'))
            # Make the file executable
            exe.chmod(exe.stat().st_mode | stat.S_IEXEC)
            log.info(f'Found {tool} at {exe}')
            paths[tool] = exe
        except StopIteration:
            pass

    def prompt(tool, source, alt_text=False):
        answer = ''
        while answer not in {'y', 'yes', 'n', 'no'}:
            if not alt_text:
                answer = input(f'Download {tool} from {source}? [y/n]').lower().strip()
            else:
                answer = input('Import paths and skip re-download? [y/n]').lower().strip()
        return answer in {'y', 'yes'}

    if other_cfg.is_file():
        other_paths = dict()
        # fetch the paths from the config
        cfg_parser = configparser.ConfigParser()
        cfg_parser.read(other_cfg)
        if 'Paths' in cfg_parser:
            other_paths.update(dict(cfg_parser['Paths']))
            if all(tool in other_paths for tool in all_tools):
                # if all three paths are there, offer to skip the re-download
                log.info(f'Found config for other AB12PHYLO version at {other_cfg}')
                if prompt('', '', True):
                    skip = True  # The user chose to not download new tools
                    paths.update(other_paths)

    for tool in all_tools:
        if skip:
            continue

        # Look for a local installation and check if it is recent enough
        exe = shutil.which(tool)
        if exe is not None:
            log.info(f'Found {tool} at {exe}')
            if tool == 'blastn':
                output = check_output(exe + ' -version', shell=True).decode('utf-8')
                version = [int(i) for i in output.split(',')[0].split(' ')[-1].split('.')]
                if version[0] == 2 and version[1] >= 9 or version[0] > 2:
                    paths[tool] = exe
                    continue
                else:
                    log.warning('installed BLAST+ is outdated')

            elif tool == 'raxml-ng':
                # check if the installed version is old-ish / the packaged version is newer
                proc = run(stdout=PIPE, stderr=PIPE, shell=True, args=f'{exe} -v')
                version = [int(i) for i in re.search(
                    'RAxML-NG v[\.\s]+?([0-9\.]+)', proc.stdout.decode()).groups()[0].split('.')]
                if version[0] == 0 or version[0] == 1 and version[1] == 0 and version[2] <= 1:
                    log.warning('installed RAxML-NG is outdated')
                else:
                    paths[tool] = exe
                    continue

        # No or outdated local installation
        # At this point, always doing a fresh download -> enables updating.
        try:
            if tool == 'blastn':
                log.info(f'{tool} not installed (not on the $PATH).')
                if not prompt('BLAST+', 'the NCBI'):
                    continue  # The user chose to not download a new BLAST+
                url = 'https://ftp.ncbi.nlm.nih.gov/blast/executables/LATEST'
                log.debug(f'Fetching {url} ... ')
                r = requests.get(url, timeout=12)
                html = BeautifulSoup(r.content, 'html.parser')
                links = [tag['href'] for tag in html.find_all('a') if 'href' in tag.attrs]

                archive_suffix = '-x64-%s.tar.gz' % {'linux': 'linux', 'win32': 'win64',
                                                     'darwin': 'macosx'}[sys.platform]

                right_one = [l for l in links if l.endswith(archive_suffix)][0]
                zf = save_dir / right_one

                log.debug(f'Downloading {right_one} ... ')
                r = requests.get(f'{url}/{right_one}', timeout=12)
                with open(zf, 'wb') as fd:
                    for chunk in r.iter_content(chunk_size=128):
                        fd.write(chunk)

                log.debug(f'Extracting BLAST+ ... ')
                with tarfile.open(zf) as zo:
                    zo.extractall(zf.parent)
                zf.unlink()
                find(tool)

            elif tool in ['raxml-ng', 'iqtree2']:
                if not prompt({'raxml-ng': 'RAxML-NG',
                               'iqtree2': 'IQ-Tree'}[tool], 'GitHub'):
                    continue

                platform = sys.platform
                url = 'https://api.github.com/repos/%s/%s/releases/latest'
                if tool == 'raxml-ng':
                    url %= 'amkozlov', 'raxml-ng'
                else:
                    url %= 'iqtree', 'iqtree2'
                log.debug(f'Fetching {url} ... ')
                js = requests.get(url, timeout=12).json()
                for asset in js['assets']:
                    name, path = asset['name'], asset['browser_download_url']
                    # find the right release
                    if not (tool == 'raxml-ng' and (platform in {'linux', 'win32'}
                                                    and 'linux_x86_64.zip' in name
                                                    or platform == 'darwin'
                                                    and 'macos_x86_64.zip' in name)
                            or tool.startswith('iqtree2') and (platform == 'linux'
                                                               and 'Linux.tar.gz' in name
                                                               or platform == 'win32'
                                                               and 'Windows.zip' in name
                                                               or platform == 'darwin'
                                                               and 'MacOSX.zip' in name)):
                        continue
                        # will not stop and therefore not download anything
                        # for e.g. RAxML-NG on Windows
                    # download
                    log.debug(f'Downloading {name} ... ')
                    r = requests.get(path, stream=True)
                    zf = save_dir / name
                    with open(zf, 'wb') as file:
                        for chunk in r.iter_content(chunk_size=128):
                            file.write(chunk)

                    # extract
                    log.debug(f'Extracting {zf} ... ')
                    zs = zf.with_suffix('')
                    if zf.suffix == '.zip':
                        with zipfile.ZipFile(zf, 'r') as zo:
                            # prevent ridiculous double level
                            if all(n.startswith(zs.name) for n in zo.namelist()):
                                zo.extractall(zs.parent)
                            else:
                                zo.extractall(zs)
                    elif zf.suffixes[-2:] == ['.tar', '.gz']:
                        zs = zs.with_suffix('')  # yes, again
                        with tarfile.open(zf) as zo:
                            # prevent ridiculous double level
                            if all(n.startswith(zs.name) for n in zo.getnames()):
                                zo.extractall(zs.parent)
                            else:
                                zo.extractall(zs)
                    zf.unlink()
                    find(tool)

        except requests.ConnectionError as ex:
            log.error('Couldn\'t download %s. Is the system offline?' % tool)
            log.error(ex)
        except Exception as ex:
            log.error('Downloading %s failed.' % tool)
            log.error(ex)

    # when nothing was downloaded, check if the tools are there and save paths
    for tool in all_tools:
        if tool not in paths:
            find(tool)

    with open(cfg, 'w') as fh:
        fh.write(f'''
# This file indicates that ab12phylo{suffix} has been run before
# and stores paths to some non-python external tools.
# Invoking ab12phylo{suffix} with the `--initialize` flag or 
# deleting the file will trigger a re-download of these 
# tools the next time ab12phylo{suffix} is run.\n''')
        fh.write(f'\n[Paths]')
        for k, v in paths.items():
            fh.write(f'\n{k}={v}')



import re
import sys
from argparse import Namespace
from pathlib import Path
from time import time

BASE_DIR = Path(__file__).resolve().parents[1]
TOOLS = BASE_DIR / 'ab12phylo_cmd' / 'tools'

PATHS = Namespace(**{
    'raw_msa': Path('Trim') / 'raw_msa.fasta',
    'import_msa': Path('import') / 'import_raw_msa.fasta',
    'msa': 'msa.fasta',
    'msa_anno': 'msa_annotated.fasta',
    'missing': 'missing_samples.tsv',
    'tsv': 'metadata.tsv',
    'xml': Path('BLAST') / 'local_blast+_result.xml',
    'www_xml': Path('BLAST') / 'online_blast_result.xml',
    'missing_fasta': Path('BLAST') / 'missing.fasta',
    'bad_seqs': 'bad_seqs.tsv',
    'preview': Path('Trim') / 'trim_preview.png',
    'cbar': Path('Trim') / 'colorbar.png',
    'gbar': Path('Trim') / 'gene_colorbar.png',
    'left': Path('Trim') / 'msa_gbl_pre.png',
    'right': Path('Trim') / 'msa_gbl_post.png',
    'phylo_msa': Path('Trim') / 'msa_phylo.png',
    'tbe': 'tree_TBE.nwk', 'tben': 'tree_TBE_annotated.nwk',
    'fbp': 'tree_FBP.nwk', 'fbpn': 'tree_FBP_annotated.nwk',
    'icon_path': BASE_DIR / 'ab12phylo' / 'files' / 'favi.png',
    'modified_tree': 'modified_tree.nwk',
    'cmd_config': BASE_DIR / 'ab12phylo_cmd' / 'config' / 'conf.cfg'
})

algos = {'MAFFT': 'mafft', 'Clustal Omega': 'clustalo', 'MUSCLE': 'muscle', 'T-Coffee': 'tcoffee',
         'RAxML-NG': 'raxml-ng', 'IQ-Tree2': 'iqtree2', 'FastTree': 'FastTree'}
toalgo = lambda c: algos[c] + ['', '.exe'][sys.platform == 'win32']
toname = lambda c: {'raxml-ng': 'RAxML-NG', 'iqtree2': 'IQ-Tree2'}.get(c, 'unknown')
NUCLEOTIDES = ['A', 'C', 'G', 'T', 'N', 'else', '-', ' ', 'S', '?']
toint_map = dict(zip(NUCLEOTIDES, range(len(NUCLEOTIDES))))
toint = lambda c: toint_map.get(c, 5)
seqtoint = lambda s: list(map(toint, s))
toseq = lambda c: NUCLEOTIDES[c]
inttoseq = lambda s: ''.join(list(map(toseq, s)))
togray_map = dict(zip(NUCLEOTIDES, seqtoint('N') * 5 + list(range(5, 9))))
togray = lambda c: togray_map.get(c, 5)
seqtogray = lambda s: list(map(togray, s))

rgx = lambda c: re.compile(r'|'.join(
    ['.*' + re.escape(word.strip()) + '.*' for word in
     c.strip().strip(',').split(',')]))

tohex = lambda c: '#' + ''.join([('0' + hex(min(255, int(round(a * 256))))[2:])[-2:].upper() for a in c])
torgba = lambda a: tuple(round(int(a[1 + i * 2: 3 + i * 2], 16) / 255, 2) for i in range(len(a) // 2))

technicolor = {
    'A': (0.92, 1, 0.4, 1),
    'C': (0.46, 1, 0.44, 1),
    'G': (0.16, 0.44, 0.8, 1),
    'T': (1, 0.47, 0.66, 1),
    'N': (0.84, 0.84, 0.84, 0.6),
    'else': (0, 0, 0, 1),
    '-': (1, 1, 1, 0),
    ' ': (1, 1, 1, 0),
    'S': (1, 1, 1, 0),
    '?': (0, 0, 0, 0)
}

green_blue = dict(zip(NUCLEOTIDES, [
    (.56, 1, .14, 1),
    (.16, .44, .8, 1),
    (.06, 1, .75, 1),  # rgb(15,255,192)
    (.92, 1, .7, 1),
    (.94, .94, .94, .6),
    (0, 0, 0, 1),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (0, 0, 0, 0)]))

viridis = dict(zip(NUCLEOTIDES, [
    (.48, .11, .44, 1),
    (.22, .55, .77, 1),
    (.1, .9, .51, 1),
    (1, .9, 0, 1),
    (.94, .94, .94, .6),
    (0, 0, 0, 1),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (0, 0, 0, 0)]))

clustalx = dict(zip(NUCLEOTIDES, [
    (.9, .2, .1, 1),  # clustal-red
    (.1, .5, .9, 1),  # clustal-blue
    (.9, .6, .3, 1),  # clustal-orange
    (.1, .8, .1, 1),  # clustal-green
    (.4, .4, .4, 1),  # clustal-dark-gray
    (0, 0, 0, 1),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (0, 0, 0, 0)]))

blue_pink = dict(zip(NUCLEOTIDES, [
    (.4, .65, .96, 1),
    (.14, .05, .8, 1),
    (.99, .33, .93, 1),
    (1, .91, .96, 1),
    (.94, .94, .94, .6),
    (0, 0, 0, 1),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (1, 1, 1, 0),
    (0, 0, 0, 0)]))

colors = list(map(tohex, map(technicolor.get, NUCLEOTIDES)))
USER = 'ab12phylo@gmail.com'
SEP = '??'
ALPHA = .25
DPI = 300
BUF_SIZE = 128 * 1024

DOWNLOAD_TIMEOUT = 3600  # an hour at most for db download

# support value colors
blue = (.2, .5, .75)
red = (1, 0, .3)
dark_red = (.8, 0, .25)
black = (.2, .2, .2)

# seaborn rocket palette
rocket = [blue] + [(0.1237, 0.0717, 0.1822), (0.2452, 0.1049, 0.2639),
                   (0.3809, 0.1206, 0.3250), (0.5172, 0.1179, 0.3545),
                   (0.6582, 0.0955, 0.3536), (0.7965, 0.1050, 0.3106),
                   (0.8949, 0.2178, 0.2531), (0.9429, 0.3754, 0.2636),
                   (0.9592, 0.5330, 0.3748), (0.9644, 0.6702, 0.5150),
                   (0.9689, 0.7980, 0.6851)]

version_regex = re.compile('\\.[\\d]+$')


def inc_priv_timestamp():
    return str(time())[6:17]


help = {0: '<b>Welcome to AB12PHYLO!</b>      To show or hide this help, press '
           '<b>F1</b> or <b>Ctrl+H</b>.\nPlease define your dataset of ABI trace '
           'files or sequence data in <i>.fasta</i> format here. For automatic '
           'mapping of reference data from different genes to the same strain, it is '
           'recommended to use GenBank <i>.fasta</i> records such as <a href="https://'
           'www.ncbi.nlm.nih.gov/nuccore/AF347033.1?report=fasta">AF347033.1</a>. '
           'If wellsplates are used in conjunction with sequence data for multiple '
           'genes, please make sure the plate layouts are identical.',
        1: 'Parse sample ID / well coordinates, gene, orientation and wellsplate ID '
           'from the file name using <a href="https://regex101.com/r/Yulwlf/8" title='
           '"development stage examples">Regular Expressions</a> with capturing groups. '
           'For <i>.fasta</i> files with multiple records, the sequence ID is used. '
           'Fields can be edited manually.',
        2: 'ABI trace files can be trimmed to remove low-quality regions here, based '
           'on phred quality scores. The maximum score is 60, indicating the probability '
           'of an incorrect base call is 1:10^6. Equivalently, 30 indicates 1:1000.',
        3: 'Create per-gene multiple sequence alignments, which are then '
           'concatenated for ML tree inference. AB12PHYLO can construct '
           'MSAs locally, use the <a href="https://www.ebi.ac.uk/Tools/msa/">'
           'EMBL-EBI</a> API or import previously created data.',
        4: 'Trim alignments using <a href="http://molevol.cmima.csic.es/castresana/'
           'Gblocks.html">Gblocks</a>. Especially for mixed datasets, consistent ends '
           'of the MSAs are important to avoid artificial grouping of sequences with '
           'similar sequencing characteristics.',
        5: 'Optionally include species annotation from a local BLAST+ in a '
           '<a href="https://ftp.ncbi.nlm.nih.gov/blast/db/">pre-compiled</a>'
           ' or custom database, or by importing XML results of a manual web '
           '<a href="https://blast.ncbi.nlm.nih.gov/Blast.cgi">BLAST</a> '
           'for one of the <i>_project_/_gene_/_gene_.fasta</i> files. '
           'Searching an online database such as <a href="'
           'https://www.ncbi.nlm.nih.gov/nucleotide/">NCBI nt</a> via the public '
           'BLAST API is also possible, but should not be the main search strategy. '
           'You can continue while BLAST is running.',
        6: 'Infer a phylogenetic tree using <a href="https://github.com/amkozlov/raxml'
           '-ng/">RAxML-NG</a> in three stages: <b>1:</b> Find the maximum-likelihood '
           'tree for the specified evolutionary model from multiple tree searches. <b>'
           '2:</b> Estimate confidence values for the phylogenetic placement of the taxa '
           'in the ML tree by Bootstrapping. <b>3:</b> '
           'Map the result of <b>2</b> to the tree from <b>1</b> as branch support values.',
        7: 'Plot the resulting ML tree and calculate basic diversity / neutrality metrics '
           'in the dataset. Click inside the visualizations to select samples, and right-'
           'click for tree operations such as rooting, dropping nodes, collapsing or '
           'extracting taxa. Species labels can be edited on the BLAST page, and shared '
           'background color indicates identical sequences.'
        }



# 2021 Leo Kaindl

import logging
import mmap
import random
import shlex
import shutil
import stat
import sys
import threading
from os import cpu_count
from pathlib import Path
from subprocess import call, run, Popen, PIPE
from time import sleep, time
from zipfile import ZipFile, ZIP_DEFLATED

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from ab12phylo import repo
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
PAGE = 6


class ml_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        iface.ml_tool.set_entry_text_column(0)
        iface.ml_tool.set_id_column(0)
        iface.ml_tool.connect('changed', self._load_tool_help)
        iface.ml_tool.connect_after('changed', lambda widget, *args: self.start_ML(widget, 'prep'))
        iface.ml_exe.connect('file-set', self._load_tool_help, True)
        iface.ultrafast.connect('toggled', self._ultrafast_hint)
        for wi in [iface.ultrafast, iface.infer_model]:
            wi.connect_after('toggled', lambda widget, *args: self.start_ML(widget, 'prep'))

        # connect the parser to basically everything
        for wi in [iface.bootstraps, iface.ml_seed, iface.cpu_use, iface.rand, iface.pars,
                   iface.evo_modify, iface.evo_model, iface.ml_help, iface.ml_cmd]:
            wi.connect('focus_out_event', self._parse_ml)

        iface.evo_model.set_model(data.evo_models)
        iface.evo_model.get_child().connect(
            'focus_out_event', self._change_evo_model, iface.evo_modify, data.ml)
        iface.evo_block = iface.evo_model.connect_after(
            'changed', self._load_model_file, iface.evo_modify)
        iface.evo_model.handler_block(iface.evo_block)

        iface.ml_run.connect('clicked', self.start_ML, 'run')
        iface.ml_export.connect('clicked', self.start_ML, 'export')
        iface.ml_import.connect('clicked', self.import_tree)

        for wi in [iface.bootstraps, iface.rand, iface.pars, iface.ml_seed]:
            wi.connect('key-press-event', self.edit_numerical_entry_up_down)
            wi.connect('changed', self.edit_numerical_entry)

        iface.ml_page_seen = False

    def _ultrafast_hint(self, widget, *args):
        """
        If the ultrafast bootstrapping button is toggled on and the current
        number of bootstraps is lower than 1000, hint that this won't work.
        """
        if not widget.get_active():
            return
        wi = self.iface.bootstraps
        val = int([i for i in [wi.get_text(), wi.get_placeholder_text()] if i][0])
        if val >= 1000:
            return
        self.show_notification(msg='ultrafast bootstrapping will only work '
                                   'with ≥ 1000 iterations', secs=10)

    def _parse_ml(self, widget, event_focus):
        """An interface parser to save changes to the UI in the model."""
        data = self.data
        iface = self.iface
        ml = data.ml
        ml.binary = iface.ml_exe.get_filename()
        for w_name in ['evo_modify', 'bootstraps', 'rand', 'pars', 'ml_seed', 'cpu_use']:
            wi = iface.__getattribute__(w_name)
            val = [i for i in [wi.get_text(), wi.get_placeholder_text()] if i][0]
            if w_name in ['bootstraps', 'rand', 'pars']:
                val = int(val)
            ml.__setattr__(w_name, val)
        ml.evo_model = data.evo_models[iface.evo_model.get_active()]
        ml.infer_model = iface.infer_model.get_active()
        ml.ultrafast = iface.ultrafast.get_active()
        tx = iface.ml_cmd.get_buffer().props.text
        if tx:
            ml.ml_cmd = tx
        if ml.evo_model[1]:
            # do not allow modifier suffixes if a model file is used
            ml.evo_modify = ''
            ml.evo_model = str(self.wd / 'RAxML' / 'user_model')
        else:
            ml.evo_model = ml.evo_model[0]

    def import_tree(self, widget):
        """Import a tree file or two"""
        dialog = Gtk.FileChooserDialog(title='import tree(s)',
                                       parent=None, select_multiple=True,
                                       action=Gtk.FileChooserAction.OPEN)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            paths = [Path(p).resolve() for p in dialog.get_filenames()]
            if len(paths) > 2:
                self.show_message_dialog('Please select at most two tree files.')
            else:
                errors = list()
                for path in paths:
                    try:
                        if 'FBP' in path.name.upper():
                            shutil.copy(path, self.wd / repo.PATHS.fbp)
                        elif 'TBE' in path.name.upper():
                            shutil.copy(path, self.wd / repo.PATHS.tbe)
                        else:
                            errors.append(path.name)
                    except shutil.SameFileError as sfe:
                        errors.append(f'{path} is already in the project')
                if errors:
                    self.show_message_dialog('Not immediately recognized as either '
                                             'tree_FBP.nwk or tree_TBE.nwk. You can also copy '
                                             'it/them to %s manually.' % self.wd, items=errors)
                else:
                    self.show_notification('imported trees:',
                                           [path.name for path in paths], 2)
                    self.set_errors(PAGE, False)

        dialog.destroy()
        self.set_changed(PAGE, False)

    @staticmethod
    def _change_evo_model(entry, event_focus, evo_modify, ml):
        combo = entry.get_parent().get_parent()
        if combo.get_active_iter():
            ml.evo_model = entry.get_text()
            evo_modify.props.sensitive = True
        else:
            tx = entry.get_text()
            combo.get_model().append([tx, None])
            ml.evo_model = tx  # save in project dataset
            LOG.debug('entered custom evo model %s' % tx)

    def _load_model_file(self, combo, evo_modify):
        if combo.get_active_iter() and combo.get_active_id() == 'from file':
            dialog = Gtk.FileChooserDialog(title='select file with partition table',
                                           parent=None, select_multiple=False,
                                           action=Gtk.FileChooserAction.OPEN)
            dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                               Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
            response = dialog.run()
            if response == Gtk.ResponseType.OK:
                tx = Path(dialog.get_filename()).resolve()
                try:
                    Path.mkdir(self.wd / 'RAxML', exist_ok=True)
                    shutil.copy(tx, self.wd / 'RAxML' / 'user_model')
                    shutil.copy(tx, self.wd / 'RAxML' / 'ml.raxml.bestModel')
                except FileNotFoundError:
                    self.show_notification('File does not exist', secs=1)
                    dialog.destroy()
                    return
                except shutil.SameFileError:
                    pass
                evo_modify.props.sensitive = False
                combo.get_model().append([tx.name, str(tx)])
                combo.set_active_id(tx.name)
                LOG.debug('selected partitioned model file: %s -> RAxML/user_model' % str(tx))
            dialog.destroy()

    def _load_tool_help(self, widget, try_path=False):
        """
        Load the RAxML-NG / IQ-Tree path from the config file or use the user-defined one.
        If --help exits with OSError, despair.
        """
        iface = self.iface
        ml = self.data.ml

        if not iface.ml_page_seen:
            iface.ml_tool.set_active_id(repo.toname(ml.tool))
        else:
            ml.tool = repo.toalgo(iface.ml_tool.get_active_text())

        for wi in [iface.iqtree_label, iface.ultrafast, iface.infer_model]:
            wi.set_sensitive(ml.tool != 'raxml-ng')

        if try_path:
            binary = widget.get_filename()
        else:
            binary = ab12phylo_app_base.CFG.get(ml.tool, shutil.which(ml.tool))

        if binary:
            # May be missing, for example no RAxML-NG on Windows
            ml.binary = binary
            iface.ml_exe.set_filename(binary)
            binary = Path(binary)
            # Ensure the file is executable, and present
            try:
                binary.chmod(binary.stat().st_mode | stat.S_IEXEC)
            except FileNotFoundError:
                LOG.error(f"File for {iface.ml_tool.get_active_text()} not found "
                          f"at {binary}. Please re-run 'ab12phylo --initialize' or "
                          f"manually correct the path in {ab12phylo_app_base.CONF}")

            try:
                res = run(args=shlex.split(f'{ml.binary} --help'),
                          stdout=PIPE, stderr=PIPE)
                iface.ml_help.props.buffer.props.text = res.stdout.decode().lstrip()
                LOG.debug('got %s --help' % ml.tool)
            except OSError:
                if sys.platform in ['win32', 'darwin', 'cygwin']:
                    self.show_notification('The selected binary doesn\'t work for this OS.\n'
                                           'Please use a different tool or export a .zip\n'
                                           'and run ML inference on a different machine.\n'
                                           + ml.binary)
                else:
                    assert sys.platform == 'linux', 'What\'s AIX?'

        try:
            # get number of available CPUs
            cpus = cpu_count()
            iface.cpu_count.set_text(str(cpus))
            cpu_adj = iface.cpu_use.get_adjustment().props
            cpu_adj.upper = cpus
            cpu_adj.value = cpus
            LOG.debug('found %d CPUs' % cpus)
        except Exception as ex:
            LOG.error('reading CPUs failed')

    def reload_ui_state(self):
        data = self.data
        iface = self.iface
        ml = data.ml
        for w_name in ['bootstraps', 'rand', 'pars']:
            iface.__getattribute__(w_name).set_text(str(ml.__getattribute__(w_name)))
        iface.ml_seed.set_text(str(ml.ml_seed) if 'ml_seed' in ml else '')
        iface.evo_model.set_active_id(ml.evo_model)
        iface.evo_modify.set_text(ml.evo_modify)
        iface.in_shell.set_active(ml.in_shell)
        iface.infer_model.set_active(ml.infer_model)
        iface.ultrafast.set_active(ml.ultrafast)

        iface.ml_tool.set_active_id(repo.toname(ml.tool))
        iface.ml_cmd.get_buffer().props.text = ml.ml_cmd
        for wi in [iface.iqtree_label, iface.ultrafast, iface.infer_model]:
            wi.set_sensitive(ml.tool != 'raxml-ng')

    def refresh(self):
        """Re-view the page. Get suggested commands for RAxML-NG and IQ-Tree"""
        LOG.debug('ML refresh')
        data = self.data
        iface = self.iface
        if not iface.ml_page_seen:
            self._load_tool_help(None)
            self.reload_ui_state(PAGE)
            iface.evo_model.handler_unblock(iface.evo_block)
            iface.evo_model.set_active_id(data.ml.evo_model)
            iface.ml_page_seen = True
            self.start_ML(None, 'prep')

        self.set_errors(PAGE, not any((self.wd / a).is_file()
                                      for a in [repo.PATHS.tbe, repo.PATHS.fbp]))
        # change the button to its default state
        (im, la), tx = iface.ml_run.get_child().get_children(), 'Run'
        im.set_from_icon_name('media-playback-start-symbolic', 4)
        la.set_text('Run')

    def start_ML(self, widget, mode, run_after=None):
        """
        Set-up the ML inference thread.
        :param widget: ignored, to enable use as callback
        :param mode: from {'prep', 'run', 'export'}
        :param run_after: for chaining calls, run after thread finishes
        """
        data = self.data
        iface = self.iface
        ml = data.ml

        if not data.genes:
            self.show_notification('No genes', secs=1)
            return
        elif iface.thread.is_alive() and mode == 'prep':
            return
        elif iface.thread.is_alive():
            # The button was pressed when it was in the 'Stop' state
            iface.pill2kill.set()
            return

        # parse interface
        # ml.tool = repo.toalgo(iface.ml_tool.get_active_text())
        self._parse_ml(None, None)

        ml.ml_seed = random.randint(0, max(1000, ml.bootstraps)) \
            if ml.ml_seed == 'random' else int(ml.ml_seed)
        iface.ml_seed.props.text = str(ml.ml_seed)
        Path.mkdir(self.wd / 'RAxML', exist_ok=True)
        Path.mkdir(self.wd / 'IQ-Tree', exist_ok=True)
        ml.in_shell = iface.in_shell.get_active()

        iface.run_after = run_after
        if ml.infer_model:
            iface.run_after = [self._unset_modelfinder]
        # to keep the progress bar up-to-date:
        ml.prev = 0
        ml.key = False
        ml.stdout = list()
        ml.seen = {'ML': set(), 'BS': set()}
        ml.motifs = {'ML': 'ML tree search #', 'BS': 'Bootstrap tree #'}
        iface.k = ml.bootstraps + ml.rand + ml.pars + 3 if mode == 'run' else 2
        if mode != 'prep':
            self.save()
            sleep(.1)

            # change the button to its stop state
            (im, la), tx = iface.ml_run.get_child().get_children(), 'Run'
            im.set_from_icon_name('media-playback-stop-symbolic', 4)
            la.set_text('Stop')
            iface.tup = (im, la, tx)

        iface.thread = threading.Thread(target=self.do_ML, args=[mode])
        GObject.timeout_add(100, self.update_ML, PAGE, ml)
        iface.thread.start()
        return  # to main loop

    def do_ML(self, mode):
        """Run the ML inference thread"""
        data = self.data
        iface = self.iface
        ml = data.ml
        start = time()
        msa = self.wd / repo.PATHS.msa
        prefix = self.wd / ('RAxML' if ml.tool == 'raxml-ng' else 'IQ-Tree')
        shell = self.wd / f"{'raxml' if ml.tool == 'raxml-ng' else 'iqtree'}_run.sh"
        errors = list()
        msg = False
        iface.i = 0
        assert ml.tool in {'raxml-ng', 'iqtree2'}, f'got unexpected tool: {ml.tool}'

        if mode == 'prep' or (mode == 'export' and ml.infer_model and ml.tool == 'iqtree2'):
            # Create the calls and write to the TextBuffer
            if ml.tool == 'raxml-ng':
                chck = '# check MSA\n' \
                       '"{raxml-ng}" --msa "{msa}" --check --model ' \
                       '"{evo_model}{model_modifier}" --prefix "{chk_prefix}"'

                inML = '# infer ML tree\n' \
                       '"{raxml-ng}" --msa "{msa}" --model ' \
                       '"{evo_model}{model_modifier}" --prefix "{ml_prefix}" ' \
                       '--seed {seed} --threads auto{{{cpus}}} --workers auto{{{cpus}}} ' \
                       '--redo --tree {start_trees}'

                boot = '# bootstrapping\n' \
                       '"{raxml-ng}" --bootstrap --msa "{msa}" --model "{inferred_model}" ' \
                       '--tree "{ml_tree}" --prefix "{bs_prefix}" --bs-trees {bootstraps} ' \
                       '--seed {seed} --threads auto{{{cpus}}} --workers auto{{{cpus}}} --redo'

                supp = '# calc. branch support\n' \
                       '"{raxml-ng}" --support --tree "{ml_tree}" --bs-trees "{bs_trees_file}" ' \
                       '--bs-metric fbp,tbe --prefix "{sp_prefix}" ' \
                       '--threads auto{{{cpus}}} --workers auto{{{cpus}}} --redo'

                iface.ml_cmd.get_buffer().props.text = '\n\n'.join((chck, inML, boot, supp))
            elif ml.tool == 'iqtree2':
                calls = list()

                if ml.infer_model and mode != 'export':
                    calls.append('# find best model\n'
                                 '"{iqtree2}" -s "{msa}" -m MF -mtree -mset raxml '
                                 '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                                 '--prefix "{mf_prefix}" ')
                else:
                    calls.append('# infer ML tree\n'
                                 '"{iqtree2}" -s "{msa}" -m "{evo_model}{model_modifier}" '
                                 '-ninit {start_trees} -ntop {start_trees} '
                                 '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                                 '--prefix "{ml_prefix}" ')
                    if not ml.ultrafast:
                        calls.append('# non-parametric bootstrapping\n'
                                     '"{iqtree2}" -s "{msa}" -m "{evo_model}{model_modifier}" '
                                     '-te "{ml_prefix}.treefile" -b {bootstraps} '
                                     '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                                     '--prefix "{bs_prefix}" ')
                    else:
                        calls.append('# ultrafast bootstrapping\n'
                                     '"{iqtree2}" -s "{msa}" -m "{evo_model}{model_modifier}" '
                                     '-t "{ml_prefix}.treefile" -B {bootstraps} -wbtl '
                                     '--seed {seed} -nt AUTO -ntmax {cpus} -redo '
                                     '--prefix "{uf_prefix}" ')
                    calls.append('# calc. TBE branch support\n'
                                 '"{iqtree2}" -sup "{ml_prefix}.treefile" '
                                 '-t "{boot_trees}" --tbe '
                                 '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                                 '--prefix "{sp_prefix}" ')

                iface.ml_cmd.get_buffer().props.text = '\n\n'.join(calls)

            if mode == 'prep':
                GObject.idle_add(self.stop_prep, errors)
                return True

        assert mode in {'run', 'export'}, f'got unexpected mode: {mode}'

        # Read call templates from buffer
        calls, descs, limits, keys, format_args = list(), list(), list(), list(), dict()
        ml.ml_cmd = iface.ml_cmd.get_buffer().props.text
        for line in ml.ml_cmd.strip().split('\n'):
            if line.startswith('#'):
                descs.append(line.strip()[2:])
            elif line:
                calls.append(line)

        # Prepare args to fill call templates
        if ml.tool == 'raxml-ng':
            keys = [False, 'ML', 'BS', False]
            limits = [0, 1, ml.rand + ml.pars + 1, iface.k - 2]
            format_args = {'raxml-ng': ml.binary,
                           'msa': msa, 'seed': ml.ml_seed,
                           'evo_model': ml.evo_model,
                           'model_modifier': ml.evo_modify,
                           'chk_prefix': prefix / 'chk',
                           'ml_prefix': prefix / 'ml',
                           'bs_prefix': prefix / 'bs',
                           'sp_prefix': prefix / 'sp',
                           'cpus': ml.cpu_use,
                           'bootstraps': ml.bootstraps,
                           'inferred_model': prefix / 'ml.raxml.bestModel',
                           'ml_tree': prefix / 'ml.raxml.bestTree',
                           'bs_trees_file': prefix / 'bs.raxml.bootstraps',
                           'start_trees': ','.join(
                               [a for a in ['rand{%d}' % ml.rand if ml.rand > 0 else None,
                                            'pars{%d}' % ml.pars if ml.pars > 0 else None] if a])}
        elif ml.tool == 'iqtree2':
            keys = ['ML', 'BS', False]
            limits = [1, ml.rand + ml.pars, iface.k - 2]
            format_args = {'iqtree2': ml.binary,
                           'msa': msa, 'seed': ml.ml_seed,
                           'evo_model': ml.evo_model,
                           'model_modifier': ml.evo_modify,
                           'mf_prefix': prefix / 'mf',
                           'ml_prefix': prefix / 'ml',
                           'bs_prefix': prefix / 'bs',
                           'uf_prefix': prefix / 'uf',
                           'sp_prefix': prefix / 'sp',
                           'cpus': ml.cpu_use,
                           'bootstraps': ml.bootstraps,
                           'boot_trees': prefix / ('uf.ufboot' if ml.ultrafast else 'bs.boottrees'),
                           'start_trees': ml.rand + ml.pars}

        if ml.in_shell:
            with open(shell, 'w') as sh:
                sh.write('#!/bin/bash\n\n')

        # loop over the stages
        for i, (arg, desc, key, prev) in enumerate(zip(calls, descs, keys, limits)):

            if ml.in_shell and mode == 'run':
                if ml.tool == 'raxml-ng':
                    with open(shell, 'a') as sh:
                        sh.write(f'# {desc}\n')
                        # sh.write(arg.format(**format_args)
                        if i != 2:
                            sh.write(arg.format(**format_args))
                        else:
                            # special case bootstrapping:
                            fifo = Path('pipe')
                            if fifo.exists():
                                fifo.unlink()
                            bash = '''
mkfifo pipe || exit 1
(%s) > pipe &
pid=$!
echo "AB12PHYLO: Bootstrapping PID is $pid"
while read -r line; do
    echo "$line"
    if [[ "${line::12}" == "Elapsed time" ]]; then
        echo "AB12PHYLO: Finished bootstrapping, terminating process $pid to ensure it exits."
        kill -s SIGTERM $pid
        break
    fi
done < pipe
rm pipe
                            '''.strip() % (arg.format(**format_args))
                            sh.write(bash)
                        sh.write('\n\necho "AB12PHYLO: %s done"\n' % desc)
                        if i != 3:
                            sh.write('\nsleep 1s\n\n')
                        continue

                elif ml.tool == 'iqtree2':
                    with open(shell, 'a') as sh:
                        sh.write(f'# {desc}\n')
                        sh.write(arg.format(**format_args))
                        sh.write('\n\necho "AB12PHYLO: %s done"\n' % desc)
                        continue

            # running RAxML or IQ-Tree, but live rather than in shell mode
            iface.text = desc
            LOG.info(iface.text)
            ml.stdout = list()
            ml.key = key
            ml.prev = prev

            # For RAxML, the first call is the MSA check - which is quick and useful.
            if not (mode == 'export' and ml.tool == 'iqtree2'):
                # read realtime RAxML output line by line
                proc = Popen(args=shlex.split(arg.format(**format_args)),
                             stdout=PIPE, stderr=PIPE)
                while True and not iface.pill2kill.is_set():
                    line = proc.stdout.readline()
                    if proc.poll() is not None:
                        sleep(.2)
                        break
                    if line:
                        lane = line.decode().rstrip()
                        ml.stdout.append(lane)
                        LOG.debug(lane)
                        if lane.startswith('Elapsed time'):
                            break  # not parsable for iqtree2
                    else:
                        sleep(.2)
                        break

            if iface.pill2kill.is_set():
                sleep(.2)
                GObject.idle_add(self.stop_ML, errors, start)
                return True

            # check for errors
            for line in ml.stdout:
                if line.startswith('ERROR'):
                    errors.append(line)  # not parsable for iqtree2
            if errors:
                GObject.idle_add(self.stop_ML, errors, start)
                return True

            if mode == 'export':
                iface.text = 'building zip'
                LOG.debug(iface.text)
                sh = f"{'raxml' if ml.tool == 'raxml-ng' else 'iqtree'}_run.sh"
                with open(sh, 'w', newline='') as sf:
                    bash = ('''\n#!/bin/bash\n\n# Execute this script via 'bash '''
                            + ('raxml' if ml.tool == 'raxml-ng' else 'iqtree')
                            + '''_run.sh'\n
# get the number of CPUs available
cpus=$(nproc)\n
BLUE='\033[0;34m'
NC='\033[0m' # No Color
print_usage() {
  printf "Limit the number of threads/logical cores via ${BLUE}-f <number>${NC}. "\n}\n
cpu_limit=400\n
while getopts 'f:' flag; do
  case "${flag}" in
    f) cpu_limit="${OPTARG}" ;;
    *) print_usage
       exit 1 ;;\n  esac\ndone\n
# find the minimum of the CPUs allowed and available
used=$(($cpu_limit<$cpus ? $cpu_limit : $cpus))\n
printf "${BLUE}$cpus${NC} CPUs available, use at most ${BLUE}$used${NC}.\nThis will proceed in a bit, interrupt with Ctrl+C\n"
print_usage
# make binary executable
chmod +x "'''
                            + ml.tool
                            + '''"\nsleep 10s''').strip()
                    sf.write(bash)
                    if ml.tool == 'raxml-ng':
                        format_args = {'raxml-ng': './raxml-ng',
                                       'msa': 'msa.fasta', 'seed': ml.ml_seed,
                                       'evo_model': Path(ml.evo_model).name,
                                       'model_modifier': ml.evo_modify,
                                       'chk_prefix': 'chk',
                                       'ml_prefix': 'ml',
                                       'bs_prefix': 'bs',
                                       'sp_prefix': 'sp',
                                       'cpus': '$used',
                                       'bootstraps': ml.bootstraps,
                                       'inferred_model': 'ml.raxml.bestModel',
                                       'ml_tree': 'ml.raxml.bestTree',
                                       'bs_trees_file': 'bs.raxml.bootstraps',
                                       'start_trees': ','.join(
                                           [a for a in ['rand{%d}' % ml.rand if ml.rand > 0 else None,
                                                        'pars{%d}' % ml.pars if ml.pars > 0 else None] if a])}
                        comments = ['\n\n# Check MSA\n', '\n# Find best ML tree\n',
                                    '\n# Compute bootstrap iterations\n',
                                    '\n# Calculate branch support\n']
                    elif ml.tool == 'iqtree2':
                        format_args = {'iqtree2': './iqtree2',
                                       'msa': 'msa.fasta', 'seed': ml.ml_seed,
                                       'evo_model': Path(ml.evo_model).name,
                                       'model_modifier': ml.evo_modify,
                                       'mf_prefix': 'mf',
                                       'ml_prefix': 'ml',
                                       'bs_prefix': 'bs',
                                       'uf_prefix': 'uf',
                                       'sp_prefix': 'sp',
                                       'cpus': '$used',
                                       'bootstraps': ml.bootstraps,
                                       'boot_trees': 'uf.ufboot' if ml.ultrafast else 'bs.boottrees',
                                       'start_trees': ml.rand + ml.pars}
                        comments = ['\n\n# Find best ML tree\n',
                                    '\n# Compute bootstrap iterations and FBP tree\n',
                                    '\n# Calculate TBE branch support\n']
                        if ml.infer_model:
                            # enable remote model inference
                            comments.insert(0, '\n\n# Find best model\n')
                            comments[1] = comments[1][1:]

                            format_args['evo_model'] = '$evo_model'
                            format_args['model_modifier'] = ''
                            format_args['BASH_REMATCH'] = [None, '{BASH_REMATCH[1]}']

                            calls.insert(0, '''
"{iqtree2}" -s "{msa}" -m MF -mtree -mset raxml -seed {seed} -nt AUTO -ntmax {cpus} -redo --prefix "mf"\n
echo "AB12PHYLO: iqtree2 ModelFinder finished"\n
# extract the suggested best model
line_regex="Best-fit model .*?$"\n
while read p; do
    # find the right line start
    if [[ $p =~ $line_regex ]]\n    then
        # revert the line to match the end; lazy modifiers don't work
        ledom_ove=$(echo $p | rev)
        model_regex="(\S+).*"
        if [[ $ledom_ove =~ $model_regex ]]
        then
            evo_model=$(echo ${BASH_REMATCH[1]} | rev)
            echo "AB12PHYLO: extracted model '$evo_model'"
        fi\n    fi\ndone < "mf.iqtree"'''.strip())
                    for comment, line in zip(comments, calls):
                        sf.write(comment)
                        sf.write(line.format(**format_args) + '\n')
                    sf.write('\n# Copy tree files\n')
                    if ml.tool == 'raxml-ng':
                        sf.write('cp sp.raxml.supportTBE tree_TBE.nwk\n')
                        sf.write('cp sp.raxml.supportFBP tree_FBP.nwk\n')
                    elif ml.tool == 'iqtree2':
                        sf.write(f'cp {"uf" if ml.ultrafast else "bs"}.contree tree_FBP.nwk\n')
                        sf.write(f'cp sp.suptree tree_TBE.nwk\n')
                    sf.write('\n')

                sleep(.05)
                GObject.idle_add(self.export_zip, sh, msa)
                return True

        if ml.in_shell and mode == 'run':  # same condition as at **1
            shell.chmod(shell.stat().st_mode | stat.S_IEXEC)
            self.hold()
            self.win.hide()
            sleep(.1)
            Popen(['notify-send', 'AB12PHYLO', 'ML Tree Inference running in background.',
                   '-i', str(repo.PATHS.icon_path)])
            call([shell])
            sleep(.1)
            self.win.show_all()
            self.release()

        iface.text = 'copy tree files'
        iface.i = iface.k - 1
        if ml.tool == 'raxml-ng':
            shutil.copy(prefix / 'sp.raxml.supportFBP', self.wd / repo.PATHS.fbp)
            shutil.copy(prefix / 'sp.raxml.supportTBE', self.wd / repo.PATHS.tbe)
        elif ml.tool == 'iqtree2':
            if ml.infer_model:
                # extract/parse model from mf.iqtree
                with open(f'{prefix / "mf"}.iqtree') as fh:
                    s = mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ)
                    pos = s.find(b'Best-fit model')
                    if pos != -1:
                        model = s[pos + 10:pos + 190].decode().split('\n')[0].split(':')[-1].strip()
                        msg = f'Extracted model: {model}'
                        LOG.info(msg)
                        ml.evo_model = model.split('+')[0]
                        ml.evo_modify = model[len(ml.evo_model):]
                        if ml.evo_model not in data.evo_models.get_column(0):
                            data.evo_models.insert(len(data.evo_models) - 1, [ml.evo_model, None])
                        iface.evo_model.set_active_id(ml.evo_model)
                        iface.evo_modify.set_text(ml.evo_modify)
                    else:
                        errors.append(f'Could not extract model from\n{prefix / "mf"}.iqtree')
            else:
                shutil.copy(f'{prefix / ("uf" if ml.ultrafast else "bs")}.contree',
                            self.wd / repo.PATHS.fbp)
                shutil.copy(prefix / 'sp.suptree', self.wd / repo.PATHS.tbe)
        iface.text = 'idle'
        iface.frac = 1
        sleep(.1)
        GObject.idle_add(self.stop_ML, errors, start, msg)
        return True

    def stop_prep(self, errors):
        """Finish the ML prep thread"""
        iface = self.iface
        iface.thread.join()
        self.win.show_all()
        if errors:
            self.show_notification('Errors during ML prep', errors)
        iface.pill2kill.clear()
        self.refresh()
        return

    def stop_ML(self, errors, start, msg=False):
        """Finish the ML inference thread"""
        iface = self.iface
        iface.thread.join()
        self.update_ML(PAGE, self.data.ml)
        self.win.show_all()

        # change the button to its stop state
        im, la, tx = iface.tup
        im.set_from_icon_name('media-playback-start-symbolic', 4)
        la.set_text(tx)
        if errors:
            self.show_notification('Errors during ML inference', errors)
        elif iface.pill2kill.is_set():
            tx = 'stopped ML inference'
            self.show_notification(msg=tx, secs=2)
        elif time() - start > 120:
            Popen(['notify-send', 'AB12PHYLO', 'ML Tree Inference finished',
                   '-i', str(repo.PATHS.icon_path)])
            # notify = threading.Thread(target=_zenity, args=())
            # notify.start()
        else:
            tx = 'ML inference finished' if not msg else msg
            self.show_notification(tx)
        LOG.info(tx)
        iface.pill2kill.clear()
        self.refresh()
        if iface.run_after and not tx.startswith('stopped'):
            [do_func() for do_func in iface.run_after]
        self.set_changed(PAGE, False)
        return

    def _unset_modelfinder(self):
        self.iface.infer_model.set_active(False)
        self._load_tool_help(self.iface.ml_tool)

    def export_zip(self, sh, msa):
        """Finish the zip export thread"""
        data = self.data
        iface = self.iface
        ml = data.ml
        iface.thread.join()
        self.win.show_all()

        path = self.wd / f'{repo.toname(ml.tool)}_export.zip'
        dialog = Gtk.FileChooserDialog(title='export zip',
                                       parent=None, select_multiple=False,
                                       action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
        dialog.set_do_overwrite_confirmation(True)
        dialog.set_current_folder(str(path.parent))
        dialog.set_current_name(path.name)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            try:
                p = Path(dialog.get_filename()).resolve()
                with ZipFile(p, 'w', ZIP_DEFLATED) as zf:
                    if ml.evo_model.endswith('user_model'):
                        zf.write(ml.evo_model, 'user_model')
                    zf.write(ml.binary, ml.tool)
                    zf.write(msa, 'msa.fasta')
                    zf.write(sh)
                Path(sh).unlink()
            except Exception as ex:
                LOG.error(ex)
        dialog.destroy()

        # change the button to its stop state
        im, la, tx = iface.tup
        im.set_from_icon_name('media-playback-start-symbolic', 4)
        la.set_text(tx)



# 2021 Leo Kaindl

"""
Builds phylogenetic trees using raxml-ng or iqtree2, parallelized in threads.
"""

import logging
import mmap
import os
import random
import re
import shutil
import string
import subprocess
import threading
from pathlib import Path

import pandas

from ab12phylo_cmd.filter import chmod_x


def ml_build(args):
    """This is just an intermediary to keep the code in main.py a tiny bit more concise."""
    if args.ml_tool == 'iqtree2':
        return iqtree_build(args)
    else:
        return raxml_build(args)


class iqtree_build:
    """
    Builds phylogenetic tree using iqtree-2.
    https://doi.org/10.1093/molbev/msu300
    """

    def __init__(self, args):
        self.log = logging.getLogger(__name__)
        self.args = args

        if self.args.ultrafast and self.args.bootstrap < 1000:
            self.log.warning(f'The selected number of iterations ({self.args.bootstrap}) '
                             f'is too low for ultrafast bootstrapping. '
                             f'Will run 1000 iterations.')
            self.args.bootstrap = 1000

        # make a directory for iqtree2
        self._dir = Path(args.dir) / 'iqtree'
        Path.mkdir(self._dir, exist_ok=True)

        # look for iqtree binary
        try:
            self._binary = self.args.cfg.get('iqtree2', shutil.which('iqtree2'))
        except KeyError:
            self._binary = shutil.which('iqtree2')
        if self._binary is None:
            self.log.error('IQ-Tree 2 not installed')
            os._exit(1)  # this is no ordinary exit; it kills zombies, too!
        # Make sure it's executable
        chmod_x(self._binary)

    def run(self):

        calls = list()
        if self.args.findmodel:
            calls.append('# find best model\n'
                         '"{iqtree2}" -s "{msa}" -m MF -mtree -mset raxml '
                         '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                         '--prefix "{mf_prefix}" ')

        calls.append('# infer ML tree\n'
                     '"{iqtree2}" -s "{msa}" -m "{evo_model}" '
                     '-ninit {start_trees} -ntop {start_trees} '
                     '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                     '--prefix "{ml_prefix}" ')

        if not self.args.ultrafast:
            calls.append('# non-parametric bootstrapping\n'
                         '"{iqtree2}" -s "{msa}" -m "{evo_model}" '
                         '-te "{ml_prefix}.treefile" -b {bootstraps} '
                         '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                         '--prefix "{bs_prefix}" ')
        else:
            calls.append('# ultrafast bootstrapping\n'
                         '"{iqtree2}" -s "{msa}" -m "{evo_model}" '
                         '-t "{ml_prefix}.treefile" -B {bootstraps} -wbtl '
                         '--seed {seed} -nt AUTO -ntmax {cpus} -redo '
                         '--prefix "{uf_prefix}" ')

        calls.append('# calc. TBE branch support\n'
                     '"{iqtree2}" -sup "{ml_prefix}.treefile" '
                     '-t "{boot_trees}" --tbe '
                     '-seed {seed} -nt AUTO -ntmax {cpus} -redo '
                     '--prefix "{sp_prefix}" ')

        # potentially limit threads
        cpus = os.cpu_count()
        if 'max_threads' in self.args:
            cpus = min(cpus, self.args.max_threads)

        # prep formatting args
        fargs = {'iqtree2': self._binary,
                 'msa': self.args.msa,
                 'seed': self.args.seed,
                 'evo_model': self.args.evomodel,
                 'cpus': cpus,
                 'mf_prefix': self._dir / 'mf',
                 'ml_prefix': self._dir / 'ml',
                 'bs_prefix': self._dir / 'bs',
                 'uf_prefix': self._dir / 'uf',
                 'sp_prefix': self._dir / 'sp',
                 'bootstraps': self.args.bootstrap,
                 'boot_trees': self._dir / ('uf.ufboot' if self.args.ultrafast else 'bs.boottrees'),
                 'start_trees': sum(self.args.start_trees)}

        for i, (desc, call) in enumerate([c.split('\n') for c in calls]):
            self.log.info(desc[2:])
            _run_sp(call.format(**fargs))

            if self.args.findmodel and i == 0:
                with open(f'{self._dir / "mf"}.iqtree') as fh:
                    s = mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ)
                    pos = s.find(b'Best-fit model')
                    if pos != -1:
                        model = s[pos + 10:pos + 190].decode().split('\n')[0].split(':')[-1].strip()
                        msg = f'Extracted model: {model}'
                        self.log.info(msg)
                        fargs['evo_model'] = model
                    else:
                        self.log.error(f'Could not extract model from\n{self._dir / "mf"}.iqtree')
                        os._exit(1)  # this is no ordinary exit; it kills zombies, too!

        # copy trees with the right support values to parent directory
        shutil.copy(f'{self._dir / ("uf" if self.args.ultrafast else "bs")}.contree',
                    self.args.final_tree + '_FBP.nwk')
        shutil.copy(self._dir / 'sp.suptree', self.args.final_tree + '_TBE.nwk')
        self.log.info('final trees %s_[FBP,TBE].nwk' % self.args.final_tree)


class raxml_build:
    """
    Builds phylogenetic trees using raxml-ng.
    https://doi.org/10.1093/bioinformatics/btz305
    """

    def __init__(self, args):
        self.log = logging.getLogger(__name__)
        self.args = args
        self.runs = self.cpus = self._prefixes = self._best_tree = None
        self.msa = self.args.msa

        # seed RNG
        random.seed(args.seed)

        # make a directory for raxml
        self._dir = Path(args.dir) / 'raxml'
        Path.mkdir(self._dir, exist_ok=True)

        # look for raxml-ng binary
        try:
            self._binary = self.args.cfg.get('raxml-ng', shutil.which('raxml-ng'))
        except KeyError:
            self._binary = shutil.which('raxml-ng')
        if self._binary is None:
            self.log.error('RAxML-NG not installed')
            os._exit(1)  # this is no ordinary exit; it kills zombies, too!
        # Make sure it's executable
        chmod_x(self._binary)

        if self.args.ultrafast or self.args.findmodel:
            self.log.warning('IQ-Tree parameters were passed for a RAxML-NG run, will be ignored.')

    def run(self):
        """
        Calls MSA check. Then creates ML search threads and keeps track of used prefixes.
        Finds best tree topology from logs, then runs Bootstrapping threads.
        Computes support values for best topology from collated bootstrap trees.
        Finally copies supported tree to parent directory.
        :return:
        """
        # check+parse msa, determine number of parallel raxml-ng instances and cpus for each
        self.runs, self.cpus, self.msa = self._check_msa()

        # keep track of finished prefixes:
        self._prefixes = []

        # keep track of threads:
        active_threads = []

        self.log.debug('starting ML searches')
        # calculate ML tree searches share
        ml_searches = [mltype // self.runs + 1 for mltype in self.args.start_trees]

        # create+start ML search threads
        for j in range(self.runs):
            prefix = 'ml_' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
            self._prefixes.append(prefix)
            seed = random.randint(0, max(1000, self.args.bootstrap))
            thread = raxml_thread((prefix, self.log, self._binary, self.msa, ml_searches[0],
                                   ml_searches[1], self.args.evomodel, self.args.min_dist, seed,
                                   self._dir / prefix, self.cpus), mode='infer_topology')
            active_threads.append(thread)
            thread.start()
        self.log.info('active Threads: %s' % ' '.join(self._prefixes))

        # finish threads
        for thread in active_threads:
            thread.join()
            self.log.info('finished %s' % thread.prefix)
        self.log.debug('all ML searches done')

        # find best topology: parse logs for best LogLikelihood
        best_prefix, best_score = '', float('-inf')
        for prefix in self._prefixes:
            with open(f'{self._dir / prefix}.raxml.log', 'r') as log:
                text = log.read()
                pos = text.find('Final LogLikelihood:')
                score = float(text[pos:pos + 40].split('\n')[0].split(' ')[-1])
                if score > best_score:
                    best_prefix, best_score = prefix, score
        self.log.info('best topology %s %.2f' % (best_prefix, best_score))

        # set to file path
        self._best_tree = f'{self._dir / best_prefix}.raxml.bestTree'

        self.log.debug('starting Bootstrapping')
        self._prefixes = []
        active_threads = []
        # calculate bootstraps share
        trees = self.args.bootstrap // self.runs + 1

        # create+start Bootstrapping threads
        for j in range(self.runs):
            prefix = 'bs_' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=4))
            self._prefixes.append(prefix)
            seed = random.randint(0, max(1000, self.args.bootstrap))
            thread = raxml_thread((prefix, self.log, self._binary, self.msa, self._best_tree,
                                   trees, self.args.evomodel, self.args.min_dist, seed,
                                   self._dir / prefix, self.cpus), mode='bootstrap')
            active_threads.append(thread)
            thread.start()
        self.log.info('active BS Threads: %s' % ' '.join(self._prefixes))

        # finish threads
        for thread in active_threads:
            thread.join()
            self.log.info('finished %s' % thread.prefix)

        self.log.debug('computing supports')
        # concat all bootstrap trees
        all_bs_trees = self._dir / 'all_bs_trees.nwk'
        with open(all_bs_trees, 'w') as fh:
            for prefix in self._prefixes:
                try:
                    with open(f'{self._dir / prefix}.raxml.bootstraps', 'r') as bootstraps:
                        fh.write(bootstraps.read() + '\n')
                except FileNotFoundError:
                    self.log.warning('results from %s missing' % prefix)
        self.log.debug('all bootstrap trees in %s' % all_bs_trees)

        run_cmd = f'{self._binary} --support --tree "{self._best_tree}" ' \
                  f'--bs-trees "{all_bs_trees}" --prefix "{self._dir / "_sup"}" ' \
                  f'--threads {self.cpus} --bs-metric fbp,tbe --redo'
        _run_sp(run_cmd)

        # copy best trees with the right support values to parent directory
        shutil.move(self._dir / '_sup.raxml.supportFBP', self.args.final_tree + '_FBP.nwk')
        shutil.move(self._dir / '_sup.raxml.supportTBE', self.args.final_tree + '_TBE.nwk')
        self.log.info('final trees %s_[FBP,TBE].nwk' % self.args.final_tree)

    def _check_msa(self):
        """
        Checks if the MSA can be understood by raxml-ng,
        parses it to binary and estimates effective number of threads.
        If replacing is off, the original MSA will be used.
        :return: number of parallel instances possible on this machine,
        number of CPUs recommended by raxml-ng for each one, and new path to MSA.
        """
        # prep paths for checking, reducing and parsing MSA
        paths = [self._dir / prefix for prefix in ['_chk', '_red']]

        # check msa
        arg = '%s --msa "%s" --check --model %s --prefix "%s"' \
              % (self._binary, self.msa, self.args.evomodel, paths[0])

        res = _run_sp(arg)

        keepers = set()
        if 'replace' in self.args and self.args.replace:
            # replace mode: replace sets of identical sequences by one representative. off by default
            # parse MSA check result for success and identical sequences
            pattern = re.compile(r'quences (.+?) and (.+?) are exactly identi')
            goners = 0

            gene = self.args.genes[0]
            df = pandas.read_csv(self.args.tsv, sep='\t', dtype={'id': str})
            df.set_index('id', inplace=True)
            df['replaces'] = ''
            df['replaced_by'] = ''

            # make temporary sub_df
            df1 = df[df.gene == gene]

            for keeper, goner in re.findall(pattern, res):
                goners += 1
                keepers.add(keeper)
                if df1.at[keeper, 'replaces'] == '':
                    df1.at[keeper, 'replaces'] = goner
                else:
                    df1.at[keeper, 'replaces'] = df1.at[keeper, 'replaces'] + ', ' + goner
                df1.at[goner, 'replaced_by'] = keeper

            # re-merge data frames
            df[df.gene == gene] = df1
            # swap columns
            cols = df.columns.to_list()

            # write to TSV
            df.to_csv(self.args.tsv, sep='\t', na_rep='', header=True, index=True)
            if goners > 0:
                self.log.warning('Identical sequences were found, %d replacements occurred' % goners)

        # look if MSA check was ok
        if 'successfully' not in res:
            self.log.error(f'error in MSA check, see {paths[0]}.raxml.log')
            exit(1)
        self.log.info('MSA check ok')

        # if any duplicates were found, a reduced alignment was written. ignore if not in replace mode
        msa = f'{paths[0]}.raxml.reduced.phy' if len(keepers) > 0 else self.msa

        # now parse, and use reduced alignment if duplicates found. resulting .rba will be smaller
        arg = f'{self._binary} --msa "{msa}" --parse --model {self.args.evomodel} --prefix "{paths[1]}"'
        res = _run_sp(arg)

        # find number of threads:
        pos = res.find('* Recommended number of threads')
        cpus = int(res[pos:pos + 60].split('\n')[0].split(' ')[-1])

        # how many parallel runs on how many cores?
        runs = max(1, os.cpu_count() // cpus)
        # potentially limit threads
        if 'max_threads' in self.args:
            runs = min(runs, self.args.max_threads)
        self.log.info('use %d cpu%s per raxml-ng. found %d physical cores. create %d instances'
                      % (cpus, '' if cpus == 1 else 's', os.cpu_count(), runs))

        if 'replace' in self.args and self.args.replace:
            return runs, cpus, f'{paths[1]}.raxml.rba'
        else:
            # if not in replace mode, just return the unparsed alignment
            return runs, cpus, self.msa


def _run_sp(cmd):
    """
    Runs a command in a subprocess. Exceptions will be caught and logged, but cause exit.
    :param cmd: command that will be run
    :return: None
    """
    log = logging.getLogger(__name__)
    log.debug(cmd)
    try:
        return subprocess.check_output(cmd, shell=True).decode('utf-8')
    except subprocess.CalledProcessError as e:
        if e.returncode == -2:
            log.exception('ML user interrupt')
        else:
            log.exception('ML process failed, returned %s\n' % str(e.returncode)
                          + e.output.decode('utf-8') if e.output is not None else '')
        os._exit(1)  # this is no ordinary exit; it kills zombies, too!


class raxml_thread(threading.Thread):
    """A thread that creates raxml-ng trees"""

    def __init__(self, args, mode):
        threading.Thread.__init__(self)
        self.prefix = args[0]
        self.log = args[1]
        if mode == 'infer_topology':
            self.run_cmd = '%s --msa "%s" --tree rand{%d},pars{%d} --model %s --blmin %s ' \
                           '--seed %d --prefix "%s" --threads %d --redo' % args[2:]
        elif mode == 'bootstrap':
            self.run_cmd = '%s --bootstrap --msa "%s" --tree "%s" --bs-trees autoMRE{%d} ' \
                           '--model "%s" --blmin %s  ' \
                           '--seed %d --prefix "%s" --threads %d --redo' % args[2:]
        else:
            self.log.error('wrong raxml_thread mode')
            os._exit(1)  # this is no ordinary exit; it kills zombies, too!

    def run(self):
        try:
            subprocess.check_output(self.run_cmd, shell=True)
        except subprocess.CalledProcessError as e:
            self.log.error('RAxML thread %s failed' % self.prefix)



# 2021 Leo Kaindl

"""
This module reads sequencing data to memory using the :class:`its_io.reader`,
does some quality control using :class:`filter`, and prepares input files for
the :class:`blast` and :class:`msa` modules with :class:`its_io.writer`.
"""

import os
import re
import pandas
import string
import random
import logging

from Bio import SeqIO
from os import path

from ab12phylo_cmd import filter


class reader:
    """
    The File Reader for :mod:`ABI` trace files in :code:`.ab1` format and :code:`.csv`
    tables of well-to-isolate coordinates. Generally reads in all :code:`.ab1` files in
    the provided directory, but accepts whitelist files for subsetting on sample IDs as
    well as files. Does some trimming, quality control and renaming using :class:`filter`.
    Sequences are saved in the per-gene dictionary :code:`self.seqdata` and source file
    names in :code:`self.metadata`. Reference sequences are renamed to avoid issues with
    :mod:`Gblocks` or the :class:`msa` tools. Sets :code:`args.genes` to the found genes
    if it was :code:`None`, favouring :code:`ITS1F`.

    :param args: :class:`argparse.Namespace` containing CLI and config information
    :returns None: object attributes are directly accessed
    """

    def __init__(self, args):
        """
        Reads in ABI trace files and .csv tables of well-to-isolate-coordinates
        using the arguments supplied to the package.

        :param args: argparse.Namespace containing CLI and config information
        :returns None: object attributes are directly accessed
        """
        self.log = logging.getLogger(__name__)
        self.args = args
        self.csvs, self.seqdata, self.metadata = dict(), dict(), dict()

        self._get_ids()
        self._get_seqs()

        # for empty genes list, use the ones that were read in
        if self.args.genes is None:
            self.args.genes = list(self.seqdata.keys())

            # prefer ITS as guiding gene. no ref-reordering necessary because no in by_order case
            if 'ITS1F' in self.args.genes:
                self.args.genes.insert(0, self.args.genes.pop(args.genes.index('ITS1F')))

        self.log.info('--GENES-- %s' % '::'.join(self.args.genes))  # leave it alone!

        self._get_refs()
        self.seq_counts = {gene: len(records) for gene, records in self.seqdata.items()}

    def _get_ids(self):
        """
        Reads in well-to-isolate coordinates from all .csv files in :param:`csv_dir`
        as :class:`pandas.DataFrame` objects. Box numbers are extracted from filenames.
        """
        if not self.args.csv_dir:
            self.args.csv_dir = os.getcwd()
            self.log.warning('searching for wellsplates in current working directory: %s' % self.args.csv_dir)
        elif self.args.csv_dir == 'ignore':
            self.log.warning('-csv argument is no valid directory. Running AB12PHYLO without wellsplates.')
            return

        pattern = re.compile(self.args.regex_csv)

        for root, dirs, files in os.walk(self.args.csv_dir):
            for file in files:
                if file.endswith('.csv'):
                    df = pandas.read_csv(path.join(root, file), header=None)
                    df.index = list(range(1, df.shape[0] + 1))
                    df.columns = list(string.ascii_uppercase[0:df.shape[1]])
                    box = pattern.search(str(file)).groups()[0]
                    if box in self.csvs:
                        self.log.error('wellsplate %s already read in. overwrite with %s' % (box, file))
                    self.csvs[box] = df
        if len(self.csvs) == 0:
            self.log.warning('No .csv files in directory. Running AB12PHYLO without wellsplates.')
        self.log.debug('read %d .csv files in: %s' % (len(self.csvs), self.args.csv_dir))
        return

    def _get_seqs(self):
        """
        Reads in all .ab1 trace files in or below :param:`args.abi_dir` as
        :class:`Bio.SeqRecord` objects and renames to sample IDs from :param:`csvs`.
        Keeps track of source files, trims sequences and marks low-quality regions using
        filter.py module. Skips seq if file parsing or coordinate extraction fails, or
        sequence has too low quality. Versionize if ID is duplicated.
        """
        if not self.args.abi_dir:
            self.args.abi_dir = os.getcwd()
            self.log.debug('searching for trace files in current working directory: %s' % self.args.abi_dir)

        # Prepare subset analysis via file or sample ID
        file_subsetting = sample_subsetting = False
        abi_whitelist = list()
        if self.args.abi_set is not None:
            file_subsetting = True
            abi_whitelist = open(self.args.abi_set, 'r').read().strip().split('\n')
            # make absolute paths
            abi_whitelist = [path.join(path.dirname(path.dirname(__file__)), abi.strip()[1:])
                             if abi != '' and abi[0] == '$' else abi.strip() for abi in abi_whitelist]
            self.log.debug('read ABI trace file whitelist with %d entries' % len(abi_whitelist))

        sample_whitelist = list()
        if self.args.sample_set is not None:
            sample_subsetting = True
            sample_whitelist = open(self.args.sample_set, 'r').read().strip().split('\n')
            self.log.info('read sample ID whitelist with %d entries' % len(sample_whitelist))

        count = 0
        # differentiate between one-regex-to-rule-them-all or three
        if self.args.regex_abi:
            pattern = re.compile(self.args.regex_abi)
        else:
            pattern = [re.compile(regex) for regex in self.args.regex_3]
        # check if we're on the lookout for reverse reads
        reverse_reads = 0
        if self.args.regex_rev:
            self.args.regex_rev = re.compile(self.args.regex_rev)
        else:
            reverse_read = False

        found_numerical = False
        with open(self.args.bad_seqs, 'w') as bad_seqs:
            bad_seqs.write('file\tid\tbox\tgene\tproblem\n')

            for root, dirs, files in os.walk(self.args.abi_dir):
                for file in files:
                    if file.endswith('.ab1'):

                        # skip files not listed if using a restricted set
                        if file_subsetting and path.join(root, file) not in abi_whitelist:
                            continue

                        count += 1

                        # read ABI file
                        try:
                            record = SeqIO.read(path.join(root, file), 'abi')
                        except UnicodeDecodeError:
                            bad_seqs.write('%s\t \t \t \tSeqIO error\n' % file)
                            self.log.error('SeqIO error: %s' % file)
                            continue

                        # parse box, gene and coordinates from name
                        # NOTE: ID of a SeqRecord sometimes indicates wrong box. -> use name.
                        try:
                            if self.args.regex_abi:
                                # single regex
                                m = pattern.search(record.name)
                                box, gene, coords = m.groups()
                            else:
                                # 3 regex
                                box = pattern[0].search(record.name).groups()[0]
                                gene = pattern[1].search(record.name).groups()[0]
                                coords = pattern[2].search(record.name).groups()[0]

                            # check if read is reverse read
                            if self.args.regex_rev:
                                reverse_read = True if re.search(self.args.regex_rev, record.name) else False
                                reverse_reads += reverse_read

                            try:
                                (row, col) = (int(coords[1:]), coords[0])
                                # swap out well coordinates for isolate numbers
                                record.id = self.csvs[box].loc[row, col]
                            except (KeyError, ValueError):
                                if len(self.csvs) == 0:
                                    # record.id = record.name.replace(gene, '')
                                    if box in ['', ' ', '-', '_']:
                                        record.id = coords.upper()
                                    else:
                                        record.id = box + '_' + coords
                                else:
                                    bad_seqs.write('%s\t \t%s\t%s\tbox not found\n' % (file, box, gene))
                                    self.log.warning('SeqIO box %s not found %s' % (box, file))

                            if record.id.isdigit():
                                record.id = f'T{record.id}'  # prepend a T so toytree won't get confused
                                found_numerical = True
                            # filter with whitelist
                            if sample_subsetting and record.id not in sample_whitelist:
                                continue

                            # skip seqs from wrong gene
                            if self.args.genes is not None and gene not in self.args.genes:
                                bad_seqs.write('%s\t%s\t%s\t%s\twrong gene\n' % (file, record.id, box, gene))
                                self.log.debug('SeqIO wrong gene %s' % file)
                                continue

                        except AttributeError:
                            bad_seqs.write('%s\t \t \t \tunexpected filename\n' % file)
                            self.log.error('SeqIO unexpected name of ABI record %s' % file)
                            continue

                        attributes = {'file': path.join(root, file), 'box': box}
                        # quality check
                        try:
                            record = filter.trim_ends(record, self.args.min_phred, self.args.end_ratio)
                            record = filter.mark_bad_stretches(record, self.args.min_phred, self.args.bad_stretch)
                            # accept reverse reads
                            if reverse_read:
                                record = record.reverse_complement(record.id, description='')
                                attributes['direction'] = 'reverse'
                                bad_seqs.write('%s\t%s\t%s\t%s\treverse read\n' % (file, record.id, box, gene))
                                self.log.debug('reverse read %s: %s' % (record.id, file))
                        except (AttributeError, ValueError) as v:
                            bad_seqs.write('%s\t%s\t%s\t%s\t%s\n' % (file, record.id, box, gene, v))
                            self.log.info('%s: %s' % (v, file))
                            if type(v) is AttributeError:
                                attributes['quality'] = str(v)
                            else:
                                continue

                        # ensure gene dict is present
                        if gene not in self.seqdata:
                            self.seqdata[gene] = dict()
                            self.metadata[gene] = dict()

                        elif record.id in self.seqdata[gene]:
                            old_id = record.id
                            # add suffix to duplicate IDs
                            record.id = filter.new_id(record.id, self.seqdata[gene].keys())
                            bad_seqs.write('%s\t%s\t%s\t%s\tnew id, %s already present\n'
                                           % (file, record.id, box, gene, old_id))
                            self.log.debug('version id %s: %s' % (record.id, file))

                        # save SeqRecord by position
                        self.seqdata[gene][record.id] = record
                        self.metadata[gene][record.id] = attributes

        if count == 0:
            self.log.error('No .ab1 ABI trace files found.')
            exit(1)
        if found_numerical:
            self.log.warning('Found records with purely numerical IDs; prepended a T to those.')
        if reverse_reads > 0:
            self.log.debug('Found %d reverse reads.' % reverse_reads)
        self.log.debug('Found %d .ab1 files in: %s' % (count, self.args.abi_dir))
        self.log.debug('Read %d sequences of acceptable quality' %
                       sum(len(gene_records) for gene_records in self.seqdata.values()))
        return

    def _get_refs(self):
        """
        Reads in reference sequences from :code:`.fasta` reference file to guiding gene in
        the per-gene-per-box dictionary of seqs. IDs are replaced by short, random ids and
        original descriptions are saved in the :code:`metadata` dictionary.
        """
        if not self.args.ref or len(self.args.ref) == 0:
            self.log.warning('Running AB12PHYLO without references.')
            return

        # use random but reproducible prefixes. different seed from raxml -> less confusion
        random.seed(self.args.seed + 2)

        # matching reference files by name to genes
        if not self.args.by_order:
            message = 'Trying to match genes and reference file names. If the allocation ' \
                      'is wrong, try setting --ref and --genes manually.'
            # convert all genes to lowercase
            genes = [gene.lower() for gene in self.args.genes]
            # extract filenames from refs
            extracts = [ref[:-6].split('/')[-1].lower() for ref in self.args.ref]

            order = list()
            for filename in extracts:
                try:
                    order.append(genes.index(filename))
                except ValueError:
                    order.append(-1)
        else:
            message = 'Assigning references by order. If mismatched, re-order --ref arguments accordingly.'
            order = list(range(len(self.args.ref)))

        self._warn_order(message, order)

        # dict of reference source organism -> random key
        lookup = dict()

        # read in refs
        for ref_file, gene_pos in zip(self.args.ref, order):
            # skip if ref file does not match any gene
            if gene_pos == -1:
                continue
            # without skip, -1 just gets the last and reads to many references
            gene = self.args.genes[gene_pos]
            count = 0

            for record in SeqIO.parse(ref_file, 'fasta'):
                count += 1
                # parse species and possibly strain
                strain = re.split(r'[\s_]', record.description.strip().split(',')[0])
                accession = strain.pop(0)
                species = strain.pop(0) + ' ' + strain.pop(0)
                try:
                    ix = strain.index('strain')
                    strain = ' '.join(strain[ix + 1:ix + 3])
                    # cope with longer composite species names
                    if ix > 0:
                        species += ' ' + ' '.join(strain[0:ix])
                except ValueError:
                    strain = species

                # catch some illegal weird html characters
                for char in ['<', '>', '\'', '"', '&']:
                    strain = strain.replace(char, '')

                # retrieve or generate random key
                if strain in lookup:
                    _id = lookup[strain]
                else:
                    # swap original ids for random short ones that no tool takes offense at.
                    _id = 'REF_' + ''.join(random.choices(string.ascii_uppercase + string.digits, k=5))
                    lookup[strain] = _id

                # save original id+description
                self.metadata[gene][_id] = \
                    {'file': ref_file, 'accession': accession, 'reference_species':
                        (species if strain == species else species + ' strain ' + strain)}
                record.id = _id
                record.description = ''  # MARK do not delete deletion
                # save ref record
                self.seqdata[gene][_id] = record

            self.log.debug('read %d ref seqs for %s from %s'
                           % (count, gene, ref_file))
        return

    def _warn_order(self, message, order):
        """
        Builds a table of gene <- ref allocations. If there are indications of errors, logs as warning.
        :param order: positions of genes that references were matched to
        """
        table = 'allocation table\n\t\t:gene:\t<-\t:reference:'
        for j in range(len(self.args.genes)):
            table += '\n\t\t%s\t<-\t' % self.args.genes[j]
            try:
                # the reference where the order is j
                table += self.args.ref[order.index(j)]
            except ValueError:
                table += '___no ref___'

        # log according to error indications
        if self.args.by_order and len(order) > 1 or -1 in order or len(order) != len(self.args.genes):
            self.log.warning(message)
            self.log.warning(table)
        else:
            self.log.info(message)
            self.log.info(table)

        # pass over order anyway
        for i in range(len(order)):
            if order[i] == -1:
                self.log.warning('Reference file could not be matched to a gene: %s' % self.args.ref[i])

        # pass over genes
        for j in range(len(self.args.genes)):
            if j not in order:
                self.log.warning('Gene %s has no reference.' % self.args.genes[j])
        return


class writer:
    """
    The :mod:`FASTA` File Writer. For each gene, :code:`<results>/<gene>/<gene>.fasta`
    will contain all sequences for the :class:`blast` and :class:`msa` modules.

    :param args: :class:`argparse.Namespace` containing CLI and config information
    :param _reader: :class:`its_io.reader` with seq and meta data
    :returns None:
    """

    def __init__(self, args, _reader):
        self.log = logging.getLogger(__name__)

        for gene in _reader.seqdata:
            # make a subdir for <gene> data
            os.makedirs(path.join(args.dir, gene), exist_ok=True)

            # write FASTA
            with open(path.join(args.dir, gene, gene + '.fasta'), 'w') as fasta:
                SeqIO.write(_reader.seqdata[gene].values(), fasta, 'fasta')
                self.log.debug('wrote %s seqdata %s' % (gene, fasta.name))

        # make a pandas DataFrame from metadata
        df = pandas.concat({gene: pandas.DataFrame.from_dict(records, orient='index')
                            for gene, records in _reader.metadata.items()})
        df.reset_index(level=[0], inplace=True)
        df.columns = ['gene'] + list(df.columns[1:])
        df.index.name = 'id'

        # write to TSV
        df.to_csv(args.tsv, sep='\t', na_rep='', header=True, index=True)
        self.log.debug('wrote metadata to %s' % args.tsv)
